#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_workplace."""

import abc
import datetime
import uuid
import warnings

import matplotlib.pyplot as plt

import plotly.figure_factory as ff
import plotly.graph_objects as go

from .base_facility import BaseFacilityState


class BaseWorkplace(object, metaclass=abc.ABCMeta):
    """BaseWorkplace.

    BaseWorkplace class for expressing workplace including facilities in a project.
    This class will be used as template.

    Args:
        name (str):
            Basic parameter.
            Name of this workplace.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None -> str(uuid.uuid4()).
        facility_list (List[BaseFacility], optional):
            Basic parameter.
            List of BaseFacility who belong to this workplace.
            Defaults to None -> [].
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted BaseTasks.
            Defaults to None -> [].
        parent_workplace (BaseWorkplace, optional):
            Basic parameter.
            Parent workplace of this workplace.
            Defaults to None.
        max_space_size (float, optional):
            Basic parameter
            Max size of space for placing components
            Default to None -> 1.0
        input_workplace_list (List[BaseWorkplace], optional):
            Basic parameter.
            List of input BaseWorkplace.
            Defaults to None -> [].
        output_workplace_list (List[BaseWorkplace], optional):
            Basic parameter.
            List of input BaseWorkplace.
            Defaults to None -> [].
        placed_component_list (List[BaseComponent], optional):
            Basic variable.
            Components which places to this workplace in simulation.
            Defaults to None -> [].
        placed_component_id_record(List[List[str]], optional):
            Basic variable.
            Record of placed components ID in simulation.
            Defaults to None -> [].
        cost_list (List[float], optional):
            Basic variable.
            History or record of this workplace's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        facility_list=None,
        targeted_task_list=None,
        parent_workplace=None,
        max_space_size=None,
        input_workplace_list=None,
        output_workplace_list=None,
        # Basic variables
        cost_list=None,
        placed_component_list=None,
        placed_component_id_record=None,
    ):
        """init."""
        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name
        self.ID = ID if ID is not None else str(uuid.uuid4())

        self.facility_list = facility_list if facility_list is not None else []
        for facility in self.facility_list:
            if facility.workplace_id is None:
                facility.workplace_id = self.ID

        self.targeted_task_list = (
            targeted_task_list if targeted_task_list is not None else []
        )
        self.parent_workplace = (
            parent_workplace if parent_workplace is not None else None
        )
        self.max_space_size = max_space_size if max_space_size is not None else 1.0

        self.input_workplace_list = (
            input_workplace_list if input_workplace_list is not None else []
        )
        self.output_workplace_list = (
            output_workplace_list if output_workplace_list is not None else []
        )

        # ----
        # Changeable variable on simulation
        # --
        # Basic variables
        if cost_list is not None:
            self.cost_list = cost_list
        else:
            self.cost_list = []

        if placed_component_list is not None:
            self.placed_component_list = placed_component_list
        else:
            self.placed_component_list = []

        if placed_component_id_record is not None:
            self.placed_component_id_record = placed_component_id_record
        else:
            self.placed_component_id_record = []

    def set_parent_workplace(self, parent_workplace):
        """
        Set `parent_workplace`.

        Args:
            parent_workplace (BaseWorkplace):
                Parent workplace
        Examples:
            >>> t = BaseWorkplace('t')
            >>> t1 = BaseWorkplace('t1')
            >>> t.set_parent_workplace(t1)
            >>> print(t.parent_workplace.name)
            't1'
        """
        self.parent_workplace = parent_workplace

    def add_facility(self, facility):
        """
        Add facility to `facility_list`.

        Args:
            facility (BaseFacility):
                Facility which is added to this workplace
        """
        facility.workplace_id = self.ID
        self.facility_list.append(facility)

    def get_total_workamount_skill(self, task_name, error_tol=1e-10):
        """
        Get total number of workamount skill of all facilities.

        By checking workamount_skill_mean_map.

        Args:
            task_name (str):
                Task name
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-10.

        Returns:
            float: total workamount skill of target task name
        """
        sum_skill_point = 0.0
        for facility in self.facility_list:
            if facility.has_workamount_skill(task_name, error_tol=error_tol):
                sum_skill_point += facility.workamount_skill_mean_map[task_name]
        return sum_skill_point

    def extend_targeted_task_list(self, targeted_task_list):
        """
        Extend the list of targeted tasks to `targeted_task_list`.

        Args:
            targeted_task_list (list[BaseTask]):
                List of targeted tasks
        Examples:
            >>> workplace = BaseWorkplace('workplace')
            >>> print([targeted_t.name for targeted_t in workplace.targeted_task_list])
            []
            >>> workplace.extend_targeted_task_list([BaseTask('t1'),BaseTask('t2')])
            >>> print([targeted_t.name for targeted_t in workplace.targeted_task_list])
            ['t1', 't2']
        """
        for targeted_task in targeted_task_list:
            self.append_targeted_task(targeted_task)

    def append_targeted_task(self, targeted_task):
        """
        Append targeted task to `targeted_task_list`.

        Args:
            targeted_task (BaseTask):
                Targeted task
        Examples:
            >>> workplace = BaseWorkplace('workplace')
            >>> print([targeted_t.name for targeted_t in workplace.targeted_task_list])
            []
            >>> t1 = BaseTask('t1')
            >>> workplace.append_targeted_task(t1)
            >>> print([targeted_t.name for targeted_t in workplace.targeted_task_list])
            ['t1']
            >>> print([target_f.name for target_f in t1.allocated_workplace_list])
            ['workplace']
        """
        self.targeted_task_list.append(targeted_task)
        targeted_task.allocated_workplace_list.append(self)

    def set_placed_component(
        self, placed_component, set_to_all_children_components=True
    ):
        """
        Set the `placed_workplace`.

        Args:
            placed_component (BaseComponent):
                Component which places to this workplace
            set_to_all_children_components (bool):
                If True, set `placed_workplace` to all children components
                Default to True
        """
        if placed_component not in self.placed_component_list:
            self.placed_component_list.append(placed_component)

            if set_to_all_children_components:
                for child_c in placed_component.child_component_list:
                    self.set_placed_component(
                        child_c,
                        set_to_all_children_components=set_to_all_children_components,
                    )

    def remove_placed_component(
        self, placed_component, remove_to_all_children_components=True
    ):
        """
        Remove the `placed_workplace`.

        Args:
            placed_component (BaseComponent):
                Component which places to this workplace
            remove_to_all_children_components (bool):
                If True, remove `placed_workplace` to all children components
                Default to True
        """
        self.placed_component_list.remove(placed_component)

        if remove_to_all_children_components:
            for child_c in placed_component.child_component_list:
                self.remove_placed_component(
                    child_c,
                    remove_to_all_children_components=remove_to_all_children_components,
                )

    def can_put(self, component, error_tol=1e-8):
        """
        Check whether the target component can be put to this workplace in this time.

        Args:
            component (BaseComponent):
                BaseComponent
            error_tol (float, optional):
                Measures against numerical error.
                Defaults to 1e-8.

        Returns:
            bool: whether the target component can be put to this workplace in this time
        """
        can_put = False
        if self.get_available_space_size() > component.space_size - error_tol:
            can_put = True
        return can_put

    def get_available_space_size(self):
        """
        Get available space size in this time.

        Returns:
            float: available space size in this time
        """
        use_space_size = sum([c.space_size for c in self.placed_component_list])
        return self.max_space_size - use_space_size

    def initialize(self, state_info=True, log_info=True):
        """
        Initialize the following changeable variables of BaseWorkplace.

        If `state_info` is True, the following attributes are initialized.

          - `placed_component_list`

        If `log_info` is True, the following attributes are initialized.
          - `cost_list`
          - `placed_component_id_record`

        BaseFacility in `facility_list` are also initialized by this function.

        Args:
            state_info (bool):
                State information are initialized or not.
                Defaluts to True.
            log_info (bool):
                Log information are initialized or not.
                Defaults to True.
        """
        if state_info:
            self.placed_component_list = []
        if log_info:
            self.cost_list = []
            self.placed_component_id_record = []
        for w in self.facility_list:
            w.initialize(state_info=state_info, log_info=log_info)

    def reverse_log_information(self, delete_head=False):
        """Reverse log information of all."""
        self.cost_list = self.cost_list[::-1]
        self.placed_component_id_record = self.placed_component_id_record[::-1]
        for facility in self.facility_list:
            facility.reverse_log_information(delete_head)
        if delete_head:
            self.cost_list.pop(0)
            # cost_head = self.cost_list.pop(0)
            # self.cost_list.append(cost_head)  # insert
            self.placed_component_id_record.pop(0)
            # log_head = self.placed_component_id_record.pop(0)
            # self.placed_component_id_record.append(log_head)  # insert

    def add_labor_cost(self, only_working=True, add_zero_to_all_facilities=False):
        """
        Add labor cost to facilities in this workplace.

        Args:
            only_working (bool, optional):
                If True, add labor cost to only WORKING facilities in this workplace.
                If False, add labor cost to all facilities in this workplace.
                Defaults to True.
            add_zero_to_all_facilities (bool, optional):
                If True, add 0 labor cost to all facilities in this workplace.
                If False, calculate labor cost normally.
                Defaults to False.

        Returns:
            float: Total labor cost of this workplace in this time.
        """
        cost_this_time = 0.0

        if add_zero_to_all_facilities:
            for facility in self.facility_list:
                facility.cost_list.append(0.0)

        else:

            if only_working:
                for facility in self.facility_list:
                    if facility.state == BaseFacilityState.WORKING:
                        facility.cost_list.append(facility.cost_per_time)
                        cost_this_time += facility.cost_per_time
                    else:
                        facility.cost_list.append(0.0)

            else:
                for facility in self.facility_list:
                    facility.cost_list.append(facility.cost_per_time)
                    cost_this_time += facility.cost_per_time

        self.cost_list.append(cost_this_time)
        return cost_this_time

    def record_assigned_task_id(self):
        """Record assigned task id."""
        for f in self.facility_list:
            f.record_assigned_task_id()

    def record_placed_component_id(self):
        """Record component id list to `placed_component_id_record`."""
        record = []
        if len(self.placed_component_list) > 0:
            # print([c for c in self.placed_component_list])
            record = [c.ID for c in self.placed_component_list]

        self.placed_component_id_record.append(record)

    def record_all_facility_state(self):
        """Record state of all facilities."""
        for facility in self.facility_list:
            facility.record_state()

    def __str__(self):
        """str.

        Returns:
            str: name of BaseWorkplace
        Examples:
            >>> workplace = BaseWorkplace("workplace")
            >>> print(workplace)
            'workplace'
        """
        return "{}".format(self.name)

    def export_dict_json_data(self):
        """
        Export the information of this workplace to JSON data.

        Returns:
            dict: JSON format data.
        """
        dict_json_data = {}
        dict_json_data.update(
            type="BaseWorkplace",
            name=self.name,
            ID=self.ID,
            facility_list=[f.export_dict_json_data() for f in self.facility_list],
            targeted_task_list=[t.ID for t in self.targeted_task_list],
            parent_workplace=self.parent_workplace.ID
            if self.parent_workplace is not None
            else None,
            max_space_size=self.max_space_size,
            cost_list=self.cost_list,
            placed_component_list=[c.ID for c in self.placed_component_list],
            placed_component_id_record=self.placed_component_id_record,
        )
        return dict_json_data

    def extract_free_facility_list(self, target_time_list):
        """
        Extract FREE facility list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract free facility from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
        Returns:
            List[BaseFacility]: List of BaseFacility
        """
        return self.__extract_state_facility_list(
            target_time_list, BaseFacilityState.FREE
        )

    def extract_working_facility_list(self, target_time_list):
        """
        Extract WORKING facility list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract working facility from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
        Returns:
            List[BaseFacility]: List of BaseFacility
        """
        return self.__extract_state_facility_list(
            target_time_list, BaseFacilityState.WORKING
        )

    def __extract_state_facility_list(self, target_time_list, target_state):
        """
        Extract state facility list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract target_state facility from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
            target_state (BaseFacilityState):
                Target state.
        Returns:
            List[BaseFacility]: List of BaseFacility
        """
        facility_list = []
        for facility in self.facility_list:
            extract_flag = True
            for time in target_time_list:
                if len(facility.state_record_list) <= time:
                    extract_flag = False
                    break
                if facility.state_record_list[time] != target_state:
                    extract_flag = False
                    break
            if extract_flag:
                facility_list.append(facility)
        return facility_list

    def get_facility_list(
        self,
        name=None,
        ID=None,
        workplace_id=None,
        cost_per_time=None,
        solo_working=None,
        workamount_skill_mean_map=None,
        workamount_skill_sd_map=None,
        state=None,
        cost_list=None,
        assigned_task_list=None,
        assigned_task_id_record=None,
    ):
        """
        Get facility list by using search conditions related to BaseFacility parameter.

        If there is no searching condition, this function returns all self.facility_list.

        Args:
            name (str, optional):
                Target facility name.
                Defaults to None.
            ID (str, optional):
                Target facility ID.
                Defaults to None.
            workplace_id (str, optional):
                Target facility workplace_id.
                Defaults to None.
            cost_per_time (float, optional):
                Target facility cost_per_time.
                Defaults to None.
            solo_working (bool, optional):
                Target facility solo_working.
                Defaults to None.
            workamount_skill_mean_map (Dict[str, float], optional):
                Target facility workamount_skill_mean_map.
                Defaults to None.
            workamount_skill_sd_map (Dict[str, float], optional):
                Target facility workamount_skill_sd_map.
                Defaults to None.
            state (BaseFacilityState, optional):
                Target facility state.
                Defaults to None.
            cost_list (List[float], optional):
                Target facility cost_list.
                Defaults to None.
            assigned_task_list (List[BaseTask], optional):
                Target facility assigned_task_list.
                Defaults to None.
            assigned_task_id_record (List[List[str]], optional):
                Target facility assigned_task_id_record.
                Defaults to None.

        Returns:
            List[BaseFacility]: List of BaseFacility.
        """
        facility_list = self.facility_list
        if name is not None:
            facility_list = list(filter(lambda x: x.name == name, facility_list))
        if ID is not None:
            facility_list = list(filter(lambda x: x.ID == ID, facility_list))
        if workplace_id is not None:
            facility_list = list(
                filter(lambda x: x.workplace_id == workplace_id, facility_list)
            )
        if cost_per_time is not None:
            facility_list = list(
                filter(lambda x: x.cost_per_time == cost_per_time, facility_list)
            )
        if solo_working is not None:
            facility_list = list(
                filter(lambda x: x.solo_working == solo_working, facility_list)
            )
        if workamount_skill_mean_map is not None:
            facility_list = list(
                filter(
                    lambda x: x.workamount_skill_mean_map == workamount_skill_mean_map,
                    facility_list,
                )
            )
        if workamount_skill_sd_map is not None:
            facility_list = list(
                filter(
                    lambda x: x.workamount_skill_sd_map == workamount_skill_sd_map,
                    facility_list,
                )
            )
        if state is not None:
            facility_list = list(filter(lambda x: x.state == state, facility_list))
        if cost_list is not None:
            facility_list = list(
                filter(lambda x: x.cost_list == cost_list, facility_list)
            )
        if assigned_task_list is not None:
            facility_list = list(
                filter(
                    lambda x: x.assigned_task_list == assigned_task_list, facility_list
                )
            )
        if assigned_task_id_record is not None:
            facility_list = list(
                filter(
                    lambda x: x.assigned_task_id_record == assigned_task_id_record,
                    facility_list,
                )
            )
        return facility_list

    def plot_simple_gantt(
        self,
        finish_margin=1.0,
        view_ready=False,
        facility_color="#D9E5FF",
        ready_color="#C0C0C0",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
    ):
        """
        Plot Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to True.
            facility_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            ready_color (str, optional):
                Ready color setting information.
                Defaults to "#C0C0C0".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
        """
        fig, gnt = self.create_simple_gantt(
            finish_margin=finish_margin,
            view_ready=view_ready,
            facility_color=facility_color,
            ready_color=ready_color,
            figsize=figsize,
            dpi=dpi,
            save_fig_path=save_fig_path,
        )
        return fig

    def create_simple_gantt(
        self,
        finish_margin=1.0,
        view_ready=False,
        facility_color="#D9E5FF",
        ready_color="#C0C0C0",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
    ):
        """
        Create Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to True.
            facility_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            ready_color (str, optional):
                Ready color setting information.
                Defaults to "#C0C0C0".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
            gnt: ax in plt.subplots()
        """
        fig, gnt = plt.subplots()
        fig.figsize = figsize
        fig.dpi = dpi
        gnt.set_xlabel("step")
        gnt.grid(True)

        target_facility_list = []
        yticklabels = []

        for facility in self.facility_list:
            target_facility_list.append(facility)
            yticklabels.append(self.name + ":" + facility.name)

        yticks = [10 * (n + 1) for n in range(len(target_facility_list))]

        gnt.set_yticks(yticks)
        gnt.set_yticklabels(yticklabels)

        for ttime in range(len(target_facility_list)):
            w = target_facility_list[ttime]
            (
                ready_time_list,
                working_time_list,
            ) = w.get_time_list_for_gannt_chart(finish_margin=finish_margin)
            if view_ready:
                gnt.broken_barh(
                    ready_time_list,
                    (yticks[ttime] - 5, 9),
                    facecolors=(ready_color),
                )
            gnt.broken_barh(
                working_time_list,
                (yticks[ttime] - 5, 9),
                facecolors=(facility_color),
            )
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig, gnt

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
        view_ready=False,
    ):
        """
        Create data for gantt plotly of BaseFacility in facility_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
        Returns:
            List[dict]: Gantt plotly information of this BaseWorkplace
        """
        df = []
        for facility in self.facility_list:
            (
                ready_time_list,
                working_time_list,
            ) = facility.get_time_list_for_gannt_chart(finish_margin=finish_margin)
            if view_ready:
                for (from_time, length) in ready_time_list:
                    to_time = from_time + length
                    df.append(
                        {
                            "Task": self.name + ": " + facility.name,
                            "Start": (
                                init_datetime + from_time * unit_timedelta
                            ).strftime("%Y-%m-%d %H:%M:%S"),
                            "Finish": (
                                init_datetime + to_time * unit_timedelta
                            ).strftime("%Y-%m-%d %H:%M:%S"),
                            "State": "READY",
                            "Type": "Facility",
                        }
                    )
            for (from_time, length) in working_time_list:
                to_time = from_time + length
                df.append(
                    {
                        "Task": self.name + ": " + facility.name,
                        "Start": (init_datetime + from_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "Finish": (init_datetime + to_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "State": "WORKING",
                        "Type": "Facility",
                    }
                )
        return df

    def create_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        save_fig_path=None,
    ):
        """
        Create Gantt chart by plotly.

        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None -> dict(WORKING="rgb(46, 137, 205)", READY="rgb(107, 127, 135)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "Type".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        colors = (
            colors
            if colors is not None
            else {"WORKING": "rgb(46, 137, 205)", "READY": "rgb(107, 127, 135)"}
        )
        index_col = index_col if index_col is not None else "State"
        df = self.create_data_for_gantt_plotly(init_datetime, unit_timedelta)
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )

        return fig

    def create_data_for_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
    ):
        """
        Create data for cost history plotly from cost_list in facility_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation

        Returns:
            data (List[go.Bar(name, x, y)]: Information of cost history chart.
        """
        data = []
        x = [
            (init_datetime + time * unit_timedelta).strftime("%Y-%m-%d %H:%M:%S")
            for time in range(len(self.cost_list))
        ]
        for facility in self.facility_list:
            data.append(go.Bar(name=facility.name, x=x, y=facility.cost_list))
        return data

    def create_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Cost Chart",
        save_fig_path=None,
    ):
        """
        Create cost chart by plotly.

        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of cost chart.
                Defaults to "Cost Chart".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        data = self.create_data_for_cost_history_plotly(init_datetime, unit_timedelta)
        fig = go.Figure(data)
        fig.update_layout(barmode="stack", title=title)
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )
        return fig

    def append_input_workplace(self, input_workplace):
        """
        Append input workplace to `input_workplace_list`.

        Args:
            input_workplace (BaseWorkplace):
                input workplace
        Examples:
           >>> workplace = BaseWorkplace("workplace")
            >>> print([input_w.name for input_w in workplace.input_workplace_list])
            []
            >>> workplace1 = BaseWorkplace("workplace1")
            >>> workplace.append_input_task(workplace1)
            >>> print([input_w.name for input_w in workplace.input_workplace_list])
            ['workplace1']
            >>> print([parent_w.name for parent_w in workplace1.output_workplace_list])
            ['workplace']
        """
        self.input_workplace_list.append(input_workplace)
        input_workplace.output_workplace_list.append(self)

    def extend_input_workplace_list(self, input_workplace_list):
        """
        Extend the list of input workplaces to `input_workplace_list`.

        Args:
            input_workplace_list (list[BaseWorkplace]):
                 List of input workplaces
        Examples:
           >>> workplace = BaseWorkplace('workplace')
            >>> print([input_w.name for input_w in workplace.input_workplace_list])
            []
            >>> workplace.extend_input_workplace_list([BaseWorkplace('wp1'),Base('wp2')])
            >>> print([input_w.name for input_t in workplace.input_workplace_list])
            ['wp1', 'wp2']
        """
        for input_workplace in input_workplace_list:
            self.input_workplace_list.append(input_workplace)
            input_workplace.output_workplace_list.append(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""organization."""

from typing import List

from .base_organization import BaseOrganization
from .base_team import BaseTeam


class Organization(BaseOrganization):
    """Organization.

    Organization class for expressing organization in target project.
    This class is implemented from BaseOrganization.

    Args:
        team_list (List[BaseTeam]):
            Basic parameter.
            List of BaseTeam in this organization.
        workplace_list (List[BaseWorkplace]):
            Basic parameter.
            List of BaseWorkplace in this organization.
            Defaults to None -> []
        cost_list (List[float], optional):
            Basic variable.
            History or record of this organization's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(self, team_list: List[BaseTeam], workplace_list=None, cost_list=None):
        """init."""
        super().__init__(team_list, workplace_list=workplace_list, cost_list=cost_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""task."""

from .base_task import BaseTask, BaseTaskState


class Task(BaseTask):
    """Task.

    Task class for expressing target workflow.
    This class is implemented from BaseTask.

    Args:
        name (str):
            Basic parameter.
            Name of this team.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None.
        default_work_amount (float, optional):
            Basic parameter.
            Defalt workamount of this BaseTask.
            Defaults to None -> 10.0.
        input_task_list (List[BaseTask], optional):
            Basic parameter.
            List of input BaseTask.
            Defaults to None -> [].
        output_task_list (List[BaseTask], optional):
            Basic parameter.
            List of output BaseTask.
            Defaults to None -> [].
        allocated_team_list (List[BaseTeam], optional):
            Basic parameter.
            List of allocated BaseTeam
            Defaults to None -> [].
        allocated_workplace_list (List[BaseWorkplace], optional):
            Basic parameter.
            List of allocated BaseWorkplace
            Defaults to None -> [].
        need_facility (bool, optional):
            Basic parameter.
            Whether one facility is needed for performing this task or not.
            Default to False
        target_component (BaseComponent, optional):
            Basic parameter.
            Target BaseComponent.
            Defaults to None.
        default_progress (float, optional):
            Basic parameter.
            Progress before starting simulation (0.0 ~ 1.0)
            Defaults to None -> 0.0.
        due_time (int, optional):
            Basic parameter.
            Defaults to None -> int(-1).
        additional_work_amount (float, optional):
            Advanced parameter.
            Defaults to None.
        auto_task (bool, optional):
            Basic parameter.
            If True, this task is performed automatically
            even if there are no allocated workers.
            Default to False.
        est (float, optional):
            Basic variable.
            Earliest start time of CPM. This will be updated step by step.
            Defaults to 0.0.
        eft (float, optional):
            Basic variable.
            Earliest finish time of CPM. This will be updated step by step.
            Defaults to 0.0.
        lst (float, optional):
            Basic variable.
            Latest start time of CPM. This will be updated step by step.
            Defaults to -1.0.
        lft (float, optional):
            Basic variable.
            Latest finish time of CPM. This will be updated step by step.
            Defaults to -1.0.
        remaining_work_amount (float, optional):
            Basic variable.
            Remaining workamount in simulation.
            Defaults to None -> default_work_amount * (1.0 - default_progress).
        state (BaseTaskState, optional):
            Basic variable.
            State of this task in simulation.
            Defaults to BaseTaskState.NONE.
        allocated_worker_list (List[BaseWorker], optional):
            Basic variable.
            State of allocating worker list in simulation.
            Defaults to None -> [].
        allocated_worker_id_record (List[List[str]], optional):
            Basic variable.
            State of allocating worker id list in simulation.
            Defaults to None -> [].
        allocated_facility_list (List[BaseFacility], optional):
            Basic variable.
            State of allocating facility list in simulation.
            Defaults to None -> [].
        allocated_facility_id_record (List[List[str]], optional):
            Basic variable.
            State of allocating facility id list in simulation.
            Defaults to None -> [].
        additional_task_flag (bool, optional):
            Advanced variable.
            Defaults to False.
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        default_work_amount=None,
        input_task_list=None,
        output_task_list=None,
        allocated_team_list=None,
        allocated_workplace_list=None,
        need_facility=False,
        target_component=None,
        default_progress=None,
        due_time=None,
        auto_task=False,
        # Advanced parameters for customized simulation
        additional_work_amount=None,
        # Basic variables
        est=0.0,
        eft=0.0,
        lst=-1.0,
        lft=-1.0,
        remaining_work_amount=None,
        state=BaseTaskState.NONE,
        allocated_worker_list=None,
        allocated_worker_id_record=None,
        allocated_facility_list=None,
        allocated_facility_id_record=None,
        # Advanced variables for customized simulation
        additional_task_flag=False,
        actual_work_amount=None,
    ):
        """init."""
        super().__init__(
            name,
            ID=ID,
            default_work_amount=default_work_amount,
            input_task_list=input_task_list,
            output_task_list=output_task_list,
            allocated_team_list=allocated_team_list,
            allocated_workplace_list=allocated_workplace_list,
            need_facility=need_facility,
            target_component=target_component,
            default_progress=default_progress,
            due_time=due_time,
            auto_task=auto_task,
            # Basic variables
            est=est,
            eft=eft,
            lst=lst,
            lft=lft,
            remaining_work_amount=remaining_work_amount,
            state=state,
            allocated_worker_list=allocated_worker_list,
            allocated_worker_id_record=allocated_worker_id_record,
            allocated_facility_list=allocated_facility_list,
            allocated_facility_id_record=allocated_facility_id_record,
        )
        # --
        # Advanced parameter for customized simulation
        self.additional_work_amount = (
            additional_work_amount if additional_work_amount is not None else 0.0
        )
        # --
        # Advanced variables for customized simulation
        if additional_task_flag is not False:
            self.additional_task_flag = additional_task_flag
        else:
            self.additional_task_flag = False
        self.actual_work_amount = self.default_work_amount * (
            1.0 - self.default_progress
        )

    def initialize(self, error_tol=1e-10, state_info=True, log_info=True):
        """
        Initialize the following changeable variables of Task.

        If `state_info` is True, the following attributes are initialized
        in addition to 'BaseTask.initialize()'.

          - additional_task_flag
          - actual_work_amount

        Args:
            state_info (bool):
                State information are initialized or not.
                Defaluts to True.
            log_info (bool):
                Log information are initialized or not.
                Defaults to True.
        """
        super().initialize(error_tol=error_tol, state_info=True, log_info=True)

        if state_info:
            self.additional_task_flag = False
            self.actual_work_amount = self.default_work_amount * (
                1.0 - self.default_progress
            )

    def perform(self, time: int, seed=None, increase_component_error=1.0):
        """
        Perform this Task in this simulation.

        @override BaseTask.perform()

        Args:
            time (int):
                Simulation time executing this method.
            seed (int, optional):
                Random seed for describing deviation of progress.
                If workamount
                Defaults to None.
            increase_component_error (float, optional):
                For advanced simulation.
                Increment error value when error has occurred.
                Defaults to 1.0.
        Note:
            This method includes advanced code of custom simulation.
            We have to separete basic code and advanced code in the future.
        """
        if self.state == BaseTaskState.WORKING:
            work_amount_progress = 0.0
            noErrorProbability = 1.0
            if self.auto_task:
                work_amount_progress = 1.0
            else:
                # for worker in self.allocated_worker_list:
                #     work_amount_progress = (
                #         work_amount_progress
                #         + worker.get_work_amount_skill_progress(self.name, seed=seed)
                #     )
                #     noErrorProbability = (
                #         noErrorProbability
                #         - worker.get_quality_skill_point(self.name, seed=seed)
                #     )
                if self.need_facility:
                    min_length = min(
                        len(self.allocated_worker_list),
                        len(self.allocated_facility_list),
                    )
                    for i in range(min_length):
                        worker = self.allocated_worker_list[i]
                        w_progress = worker.get_work_amount_skill_progress(
                            self.name, seed=seed
                        )
                        facility = self.allocated_facility_list[i]
                        f_progress = facility.get_work_amount_skill_progress(
                            self.name, seed=seed
                        )
                        work_amount_progress += w_progress * f_progress

                        noErrorProbability = (
                            noErrorProbability
                            - worker.get_quality_skill_point(self.name, seed=seed)
                        )
                else:
                    for worker in self.allocated_worker_list:
                        work_amount_progress = (
                            work_amount_progress
                            + worker.get_work_amount_skill_progress(
                                self.name, seed=seed
                            )
                        )
                        noErrorProbability = (
                            noErrorProbability
                            - worker.get_quality_skill_point(self.name, seed=seed)
                        )

            self.remaining_work_amount = (
                self.remaining_work_amount - work_amount_progress
            )

            if self.target_component is not None:
                self.target_component.update_error_value(
                    noErrorProbability, increase_component_error, seed=seed
                )
        else:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_component."""

import abc
import datetime
import uuid
from enum import IntEnum

from .base_task import BaseTaskState


class BaseComponentState(IntEnum):
    """BaseComponentState."""

    NONE = 0
    READY = 1
    WORKING = 2
    FINISHED = -1
    REMOVED = -2


class BaseComponent(object, metaclass=abc.ABCMeta):
    """BaseComponent.

    BaseComponent class for expressing target product.
    This class can be used as template.

    Args:
        name (str):
            Basic parameter.
            Name of this component.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
        parent_component_list(List[BaseComponent], optional):
            Basic parameter.
            List of parent BaseComponents.
            Defaults to None -> [].
        child_component_list (List[BaseComponent], optional):
            Basic parameter.
            List of child BaseComponents.
            Defaults to None -> [].
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted tasks.
            Defaults to None -> [].
        space_size (float, optional):
            Basic parameter.
            Space size related to base_workplace's max_space_size.
            Default to None -> 1.0.
        state (BaseComponentState, optional):
            Basic variable.
            State of this task in simulation.
            Defaults to BaseComponentState.NONE.
        state_record_list (List[BaseComponentState], optional):
            Basic variable.
            Record list of state.
            Defaults to None -> [].
        placed_workplace (BaseWorkplace, optional):
            Basic variable.
            A workplace which this componetnt is placed in simulation.
            Defaults to None.
        placed_workplace_id_record (List[str], optional):
            Basic variable.
            Record of placed workplace ID in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        parent_component_list=None,
        child_component_list=None,
        targeted_task_list=None,
        space_size=None,
        # Basic variables
        state=BaseComponentState.NONE,
        state_record_list=None,
        placed_workplace=None,
        placed_workplace_id_record=None,
    ):
        """init."""
        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.name = name
        self.ID = ID if ID is not None else str(uuid.uuid4())

        if parent_component_list is not None:
            self.parent_component_list = parent_component_list
        else:
            self.parent_component_list = []

        if child_component_list is not None:
            self.child_component_list = child_component_list
        else:
            self.child_component_list = []

        if targeted_task_list is not None:
            self.targeted_task_list = targeted_task_list
        else:
            self.targeted_task_list = []

        if space_size is not None:
            self.space_size = space_size
        else:
            self.space_size = 1.0

        if state is not BaseComponentState.NONE:
            self.state = state
        else:
            self.state = BaseComponentState.NONE

        if state_record_list is not None:
            self.state_record_list = state_record_list
        else:
            self.state_record_list = []

        if placed_workplace is not None:
            self.placed_workplace = placed_workplace
        else:
            self.placed_workplace = None

        if placed_workplace_id_record is not None:
            self.placed_workplace_id_record = placed_workplace_id_record
        else:
            self.placed_workplace_id_record = []

    def extend_child_component_list(self, child_component_list):
        """
        Extend the list of child components.

        Args:
            child_component_list (List[BaseComponent]):
                List of BaseComponents which are children of this component.
        Examples:
            >>> c = BaseComponent('c')
            >>> print([child_c.name for child_c in c.child_component_list])
            []
            >>> c.extend_child_component_list([BaseComponent('c1'),BaseComponent('c2')])
            >>> print([child_c.name for child_c in c.child_component_list])
            ['c1', 'c2']
        """
        for child_c in child_component_list:
            self.append_child_component(child_c)

    def append_child_component(self, child_component):
        """
        Append child component to `child_component_list`.

        Args:
            child_component (BaseComponent):
                BaseComponent which is child of this component.
        Examples:
            >>> c = BaseComponent('c')
            >>> print([child_c.name for child_c in c.child_component_list])
            []
            >>> c1 = BaseComponent('c1')
            >>> c.append_child_component(c1)
            >>> print([child_c.name for child_c in c.child_component_list])
            ['c1']
            >>> print([parent_c.name for parent_c in c1.parent_component_list])
            ['c']
        """
        self.child_component_list.append(child_component)
        child_component.parent_component_list.append(self)

    def set_placed_workplace(self, placed_workplace, set_to_all_children=True):
        """
        Set the `placed_workplace`.

        Args:
            placed_workplace (BaseWorkplace):
                Workplace placed in this component
            set_to_all_children (bool):
                If True, set placed_workplace to all children components
                Default to True
        """
        self.placed_workplace = placed_workplace

        if set_to_all_children:
            for child_c in self.child_component_list:
                child_c.set_placed_workplace(
                    placed_workplace, set_to_all_children=set_to_all_children
                )

    def is_ready(self):
        """
        Check READY component or not.

        READY component is defined by satisfying the following conditions:

          - All tasks are not NONE.
          - There is no WORKING task in this component.
          - The states of append_targeted_task includes READY.

        Returns:
            bool: this component is READY or not.
        """
        all_none_flag = all(
            [task.state == BaseTaskState.NONE for task in self.targeted_task_list]
        )

        any_working_flag = any(
            [task.state == BaseTaskState.WORKING for task in self.targeted_task_list]
        )

        any_ready_flag = any(
            [task.state == BaseTaskState.READY for task in self.targeted_task_list]
        )

        all_finished_flag = all(
            [task.state == BaseTaskState.FINISHED for task in self.targeted_task_list]
        )

        if all_finished_flag:
            return False

        if not all_none_flag and (not any_working_flag) and any_ready_flag:
            return True

        return False

    def extend_targeted_task_list(self, targeted_task_list):
        """
        Extend the list of targeted tasks to `targeted_task_list`.

        Args:
            targeted_task_list (List[BaseTask]):
                List of targeted tasks
        Examples:
            >>> c = BaseComponent('c')
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            []
            >>> c.extend_targeted_task_list([BaseTask('t1'),BaseTask('t2')])
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            ['t1', 't2']
        """
        for targeted_task in targeted_task_list:
            self.append_targeted_task(targeted_task)

    def append_targeted_task(self, targeted_task):
        """
        Append targeted task to `targeted_task_list`.

        Args:
            targeted_task (BaseTask):
                Targeted task of this component
        Examples:
            >>> c = BaseComponent('c')
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            []
            >>> t1 = BaseTask('t1')
            >>> c.append_targeted_task(t1)
            >>> print([targeted_t.name for targeted_t in c.targeted_task_list])
            ['t1']
            >>> print(t1.target_component.name)
            'c'
        """
        self.targeted_task_list.append(targeted_task)
        targeted_task.target_component = self

    def initialize(self, state_info=True, log_info=True, check_task_state=True):
        """
        Initialize the following changeable basic variables of BaseComponent.

        If `state_info` is True, the following attributes are initialized.

          - `state`
          - `placed_workplace`

        If `log_info` is True, the following attributes are initialized.

          - `state_record_list`
          - `placed_workplace_id_record`

        Args:
            state_info (bool):
                State information are initialized or not.
                Defaluts to True.
            log_info (bool):
                Log information are initialized or not.
                Defaults to True.
            check_task_state (bool):
                Check the state of each task in this component or not.
                Defaults to True.
        """
        if log_info:
            self.state_record_list = []
            self.placed_workplace_id_record = []

        if state_info:
            self.state = BaseComponentState.NONE
            self.placed_workplace = None

            if check_task_state:
                self.check_state()

    def reverse_log_information(self, delete_head=False):
        """Reverse log information of all."""
        self.state_record_list = self.state_record_list[::-1]
        self.placed_workplace_id_record = self.placed_workplace_id_record[::-1]
        if delete_head:
            self.state_record_list.pop(0)
            # cost_head = self.state_record_list.pop(0)
            # self.state_record_list.append(cost_head)  # insert
            self.placed_workplace_id_record.pop(0)
            # log_head = self.placed_workplace_id_record.pop(0)
            # self.placed_workplace_id_record.append(log_head)  # insert

    def check_state(self):
        """Check and update the `state` of this component."""
        self.__check_ready()
        self.__check_working()
        self.__check_finished()

    def __check_ready(self):
        if not all(
            map(lambda t: t.state == BaseTaskState.WORKING, self.targeted_task_list)
        ):
            if not all(
                map(
                    lambda t: t.state == BaseTaskState.FINISHED, self.targeted_task_list
                )
            ):
                if any(
                    map(
                        lambda t: t.state == BaseTaskState.READY,
                        self.targeted_task_list,
                    )
                ):
                    self.state = BaseComponentState.READY

    def __check_working(self):
        if any(
            map(lambda t: t.state == BaseTaskState.WORKING, self.targeted_task_list)
        ):
            self.state = BaseComponentState.WORKING

    def __check_finished(self):
        if all(
            map(lambda t: t.state == BaseTaskState.FINISHED, self.targeted_task_list)
        ):
            self.state = BaseComponentState.FINISHED

    def record_placed_workplace_id(self):
        """Record workplace id in this time to `placed_workplace_id_record`."""
        record = None
        if self.placed_workplace is not None:
            record = self.placed_workplace.ID
        self.placed_workplace_id_record.append(record)

    def record_state(self):
        """Record current `state` in `state_record_list`."""
        self.state_record_list.append(self.state)

    def __str__(self):
        """str.

        Returns:
            str: name of BaseComponent
        Examples:
            >>> c = BaseComponent("c")
            >>> print(c)
            'c'
        """
        return "{}".format(self.name)

    def export_dict_json_data(self):
        """
        Export the information of this component to JSON data.

        Returns:
            dict: JSON format data.
        """
        dict_json_data = {}
        dict_json_data.update(
            type="BaseComponent",
            name=self.name,
            ID=self.ID,
            parent_component_list=[c.ID for c in self.parent_component_list],
            child_component_list=[c.ID for c in self.child_component_list],
            targeted_task_list=[t.ID for t in self.targeted_task_list],
            space_size=self.space_size,
            state=int(self.state),
            state_record_list=[int(state) for state in self.state_record_list],
            placed_workplace=self.placed_workplace.ID
            if self.placed_workplace is not None
            else None,
            placed_workplace_id_record=self.placed_workplace_id_record,
        )
        return dict_json_data

    def get_time_list_for_gannt_chart(self, finish_margin=1.0):
        """
        Get ready/working time_list for drawing Gantt chart.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
        Returns:
            List[tuple(int, int)]: ready_time_list including start_time, length
            List[tuple(int, int)]: working_time_list including start_time, length
        """
        ready_time_list = []
        working_time_list = []
        previous_state = BaseComponentState.NONE
        from_time = -1
        to_time = -1
        for time, state in enumerate(self.state_record_list):
            if state != previous_state:
                if from_time == -1:
                    from_time = time
                elif to_time == -1:
                    to_time = time
                    if (
                        state == BaseComponentState.NONE
                        or state == BaseComponentState.FINISHED
                    ):
                        if previous_state == BaseComponentState.WORKING:
                            working_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                        elif previous_state == BaseComponentState.READY:
                            ready_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                    if state == BaseComponentState.READY:
                        if previous_state == BaseComponentState.WORKING:
                            working_time_list.append(
                                (from_time, (to_time - 1) - from_time + finish_margin)
                            )
                    if state == BaseComponentState.WORKING:
                        if previous_state == BaseComponentState.READY:
                            ready_time_list.append(
                                (from_time, to_time - from_time + finish_margin)
                            )
                    from_time = time
                    to_time = -1
            previous_state = state

            if previous_state == BaseComponentState.WORKING:
                working_time_list.append((from_time, time - from_time + finish_margin))
            elif previous_state == BaseComponentState.READY:
                ready_time_list.append((from_time, time - from_time + finish_margin))
        return ready_time_list, working_time_list

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
        view_ready=False,
    ):
        """
        Create data for gantt plotly.

        From start_time_list and finish_time_list of BaseTask in targeted_task_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.

        Returns:
            list[dict]:
                Gantt plotly information of this BaseComponent
        """
        df = []
        (
            ready_time_list,
            working_time_list,
        ) = self.get_time_list_for_gannt_chart(finish_margin=finish_margin)

        if view_ready:
            for (from_time, length) in ready_time_list:
                to_time = from_time + length
                df.append(
                    {
                        "Task": self.name,
                        "Start": (init_datetime + from_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "Finish": (init_datetime + to_time * unit_timedelta).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "State": "READY",
                        "Type": "Component",
                    }
                )
        for (from_time, length) in working_time_list:
            to_time = from_time + length
            df.append(
                {
                    "Task": self.name,
                    "Start": (init_datetime + from_time * unit_timedelta).strftime(
                        "%Y-%m-%d %H:%M:%S"
                    ),
                    "Finish": (init_datetime + to_time * unit_timedelta).strftime(
                        "%Y-%m-%d %H:%M:%S"
                    ),
                    "State": "WORKING",
                    "Type": "Component",
                }
            )
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""workflow."""

from typing import List

from .base_task import BaseTask
from .base_workflow import BaseWorkflow


class Workflow(BaseWorkflow):
    """Workflow.

    Workflow class for expressing workflow in a project.
    This class is implemented from BaseWorkflow.

    Args:
        task_list (List[BaseTask]):
            Basic parameter.
            List of BaseTask in this BaseWorkflow.
        critical_path_length (float, optional):
            Basic variable.
            Critical path length of PERT/CPM.
            Defaults to 0.0.
    """

    def __init__(self, task_list: List[BaseTask]):
        """init."""
        super().__init__(task_list)

    def perform(self, time: int, seed=None, increase_component_error=1.0):
        """
        Perform Task in task_list in simulation.

        @override BaseWorkflow.perform()

        Args:
            time (int):
                Simulation time.
            seed (int, optional):
                Random seed for describing deviation of progress.
                If workamount
                Defaults to None.
            increase_component_error (float, optional):
                For advanced simulation.
                Increment error value when error has occurred.
                Defaults to 1.0.
        Note:
            This method includes advanced code of custom simulation.
            We have to separete basic code and advanced code in the future.
        """
        for task in self.task_list:
            task.perform(
                time, seed=seed, increase_component_error=increase_component_error
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_priority_rule."""

from enum import IntEnum


class WorkplacePriorityRuleMode(IntEnum):
    """WorkplacePriorityRuleMode."""

    FSS = 0  # Free Space Size
    SSP = 1  # Sum Skill Points of targeted task


class ResourcePriorityRuleMode(IntEnum):
    """ResourcePriorityRuleMode."""

    SSP = 0
    VC = 1
    HSV = 2


class TaskPriorityRuleMode(IntEnum):
    """TaskPriorityRuleMode."""

    TSLACK = 0
    EST = 1  # Earliest Start Time
    SPT = 2  # Shortest Processing Time
    LPT = 3  # Longest Processing Time
    FIFO = 4  # First in First Out
    LRPT = 5  # Longest Remaining Process Time
    SRPT = 6  # Shortest Remaining Process Time
    LWRPT = 7  # Longest Workflow Remaining Process Time
    SWRPT = 8  # Shortest Workflow Remaining Process Time


def sort_workplace_list(
    workplace_list, priority_rule_mode=WorkplacePriorityRuleMode.FSS, **kwargs
):
    """
    Sort workplace_list as priority_rule_mode.

    Args:
        workplace_list (List[BaseWorkplace]):
            Target workplace list of sorting.
        priority_rule_mode (WorkplacePriorityRuleMode, optional):
            Mode of priority rule for sorting.
            Defaults to WorkplacePriorityRuleMode.FSS
        args:
            Other information of each rule.
    Returns:
        List[BaseWorkplace]: resource_list after sorted
    """
    # FSS: Free Space Size
    if priority_rule_mode == WorkplacePriorityRuleMode.FSS:
        workplace_list = sorted(
            workplace_list,
            key=lambda workplace: workplace.get_available_space_size(),
            reverse=True,
        )
    # SSP: Sum Skill Points of targeted task
    elif priority_rule_mode == WorkplacePriorityRuleMode.SSP:

        def count_sum_skill_point(wp, task_name):
            skill_points = sum(
                [
                    facility.workamount_skill_mean_map[task_name]
                    for facility in wp.facility_list
                    if facility.has_workamount_skill(task_name)
                ]
            )
            return skill_points

        workplace_list = sorted(
            workplace_list,
            key=lambda workplace: count_sum_skill_point(workplace, kwargs["name"]),
            reverse=True,
        )
    return workplace_list


def sort_resource_list(
    resource_list, priority_rule_mode=ResourcePriorityRuleMode.SSP, **kwargs
):
    """
    Sort resource_list as priority_rule_mode.

    Args:
        resource_list (List[BaseResource]):
            Target resource list of sorting.
            BaseWorker and BaseFacility are one of resource type in this repository.
        priority_rule_mode (ResourcePriorityRuleMode, optional):
            Mode of priority rule for sorting.
            Defaults to ResourcePriorityRuleMode.SSP
        args:
            Other information of each rule.
    Returns:
        List[BaseResource]: resource_list after sorted
    """
    # SSP: a resource which amount of skillpoint is lower has high priority
    if priority_rule_mode == ResourcePriorityRuleMode.SSP:
        resource_list = sorted(
            resource_list,
            key=lambda resource: sum(resource.workamount_skill_mean_map.values()),
        )
    # VC :a resource which cost is lower has high priority
    elif priority_rule_mode == ResourcePriorityRuleMode.VC:
        resource_list = sorted(
            resource_list,
            key=lambda resource: (resource.cost_per_time),
        )
    # HSV: a resource which target skillpoint is higher has high priority
    elif priority_rule_mode == ResourcePriorityRuleMode.HSV:
        resource_list = sorted(
            resource_list,
            key=lambda resource: resource.workamount_skill_mean_map.get(
                kwargs["name"], -float("inf")
            ),
            reverse=True,
        )

    return resource_list


def sort_task_list(task_list, priority_rule_mode=TaskPriorityRuleMode.TSLACK):
    """
    Sort task_list as priority_rule_mode.

    Args:
        task_list (List[BaseTask]):
            Target task list of sorting.
        priority_rule_mode (ResourcePriorityRuleMode, optional):
            Mode of priority rule for sorting.
            Defaults to TaskPriorityRuleMode.TSLACK
    Returns:
        List[BaseTask]: task_list after sorted
    """
    # Task: TSLACK (a task which Slack time(LS-ES) is lower has high priority)
    if priority_rule_mode == TaskPriorityRuleMode.TSLACK:
        task_list = sorted(task_list, key=lambda task: task.lst - task.est)

    elif priority_rule_mode == TaskPriorityRuleMode.EST:
        # Task: EST (a task which EST is lower has high priority)
        task_list = sorted(task_list, key=lambda task: task.est)

    elif priority_rule_mode == TaskPriorityRuleMode.SPT:
        # Task: SPT (a task which default_work_amount is lower has high priority)
        task_list = sorted(task_list, key=lambda task: task.default_work_amount)

    elif priority_rule_mode == TaskPriorityRuleMode.LPT:
        # Task: LPT (a task which default_work_amount is higher has high priority)
        task_list = sorted(
            task_list, key=lambda task: task.default_work_amount, reverse=True
        )
    elif priority_rule_mode == TaskPriorityRuleMode.FIFO:
        # Task: FIFO (First In First Out rule)
        def count_ready(x):
            k = x.state_record_list
            num = len([i for i in range(len(k)) if k[i].name == "READY"])
            return num

        task_list = sorted(task_list, key=lambda task: count_ready(task), reverse=True)
    elif priority_rule_mode == TaskPriorityRuleMode.LRPT:
        # Task: LRPT (Longest Remaining Process Time)
        task_list = sorted(
            task_list,
            key=lambda task: task.remaining_work_amount,
            reverse=True,
        )
    elif priority_rule_mode == TaskPriorityRuleMode.SRPT:
        # Task: SRPT (Shortest Remaining Process Time)
        task_list = sorted(
            task_list,
            key=lambda task: task.remaining_work_amount,
        )
    elif priority_rule_mode == TaskPriorityRuleMode.LWRPT:
        # Task: LWRPT (Longest Workflow Remaining Process Time)
        task_list = sorted(
            task_list,
            key=lambda task: task.parent_workflow.critical_path_length,
            reverse=True,
        )
    elif priority_rule_mode == TaskPriorityRuleMode.SWRPT:
        # Task: SWRPT (Shortest Workflow Remaining Process Time)
        task_list = sorted(
            task_list,
            key=lambda task: task.parent_workflow.critical_path_length,
        )

    return task_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""component."""

import numpy as np

from .base_component import BaseComponent


class Component(BaseComponent):
    """Component.

    Component class for expressing target product.
    This class is implemented from BaseComponent.

    Args:
        name (str):
            Basic parameter.
            Name of this component.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None.
        parent_component_list(List[BaseComponent], optional):
            Basic parameter.
            List of parent BaseComponents. Defaults to None.
        child_component_list (List[BaseComponent], optional):
            Basic parameter.
            List of child BaseComponents.
            Defaults to None.
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted tasks.
            Defaults to None.
        error_tolerance (float, optional):
            Advanced parameter.
        error (float, optional):
            Advanced variables.
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        parent_component_list=None,
        child_component_list=None,
        targeted_task_list=None,
        # Advanced parameters for customized simulation
        error_tolerance=None,
        # Advanced variables for customized simulation
        error=None,
    ):
        """init."""
        super().__init__(
            name,
            ID=ID,
            child_component_list=child_component_list,
            parent_component_list=parent_component_list,
            targeted_task_list=targeted_task_list,
        )
        # --
        # Advanced parameter for customized simulation
        if error_tolerance is not None:
            self.error_tolerance = error_tolerance
        else:
            self.error_tolerance = 0.0
        # Advanced variables for customized simulation
        if error is not None:
            self.error = error
        else:
            self.error = 0.0

    def initialize(self, state_info=True, log_info=True):
        """
        Initialize the following changeable variables of Component.

        If `state_info` is True, the following attributes are initialized
        in addition to 'BaseComponent.initialize()'.

          - error

        Args:
            state_info (bool):
                State information are initialized or not.
                Defaluts to True.
            log_info (bool):
                Log information are initialized or not.
                Defaults to True.
        """
        super().initialize(state_info=True, log_info=True)
        if state_info:
            self.error = 0.0

    def update_error_value(
        self, no_error_prob: float, error_increment: float, seed=None
    ):
        """
        Update error value randomly.

        If no_error_prob >=1.0, error = error + error_increment.

        Args:
            no_error_prob (float): Probability of no error (0.0~1.0)
            error_increment (float): Increment of error variables if error has occurred.
            seed (int, optional): seed of creating random.rand(). Defaults to None.
        Examples:
            >>> c = Component("c")
            >>> c.update_error_value(0.9, 1.0, seed=32)
            >>> print(c.error)
            0.0
            >>> c.update_error_value(0.4, 1.0, seed=32)
            >>> print(c.error) # Random 1.0 or 0.0
            1.0
        Note:
            This method is developed for customized simulation.
        """
        if seed is not None:
            np.random.seed(seed=seed)
        if np.random.rand() > no_error_prob:
            self.error = self.error + error_increment



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""product."""

from typing import List

from .base_component import BaseComponent
from .base_product import BaseProduct


class Product(BaseProduct):
    """BaseProduct.

    Product class for expressing target product in a project.
    This class is implemented from BaseProduct.

    Args:
        component_list (List[BaseComponent]):
            List of BaseComponents
    """

    def __init__(self, component_list: List[BaseComponent]):
        """init."""
        super().__init__(component_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""team."""

from .base_team import BaseTeam


class Team(BaseTeam):
    """Team.

    Team class for expressing team in a project.
    This class is implemented from BaseTeam.

    Args:
        name (str):
            Basic parameter.
            Name of this team.
        ID (str, optional):
            Basic parameter.
            ID will be defined automatically.
            Defaults to None.
        worker_list (List[BaseWorker], optional):
            Basic parameter.
            List of BaseWorker who belong to this team.
            Defaults to None -> [].
        targeted_task_list (List[BaseTask], optional):
            Basic parameter.
            List of targeted BaseTasks.
            Defaults to None -> [].
        parent_team (BaseTeam, optional):
            Basic parameter.
            Parent team of this team.
            Defaults to None.
        cost_list (List[float], optional):
            Basic variable.
            History or record of this team's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        name: str,
        ID=None,
        worker_list=None,
        targeted_task_list=None,
        parent_team=None,
        # Basic variables
        cost_list=None,
    ):
        """init."""
        super().__init__(
            name,
            ID=ID,
            worker_list=worker_list,
            targeted_task_list=targeted_task_list,
            parent_team=parent_team,
            cost_list=cost_list,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_workflow."""

import abc
import datetime
import warnings
from typing import List

import matplotlib.pyplot as plt

import networkx as nx

import plotly.figure_factory as ff
import plotly.graph_objects as go

from .base_facility import BaseFacilityState
from .base_task import BaseTask, BaseTaskDependency, BaseTaskState
from .base_worker import BaseWorkerState


class BaseWorkflow(object, metaclass=abc.ABCMeta):
    """BaseWorkflow.

    BaseWorkflow class for expressing workflow in a project.
    BaseWorkflow is consist of multiple BaseTasks.
    This class will be used as template.

    Args:
        task_list (List[BaseTask]):
            Basic parameter.
            List of BaseTask in this BaseWorkflow.
        critical_path_length (float, optional):
            Basic variable.
            Critical path length of PERT/CPM.
            Defaults to 0.0.
    """

    def __init__(
        self,
        # Basic parameters
        task_list: List[BaseTask],
        # Basic variables
        critical_path_length=0.0,
    ):
        """init."""
        # ----
        # Constraint parameter on simulation
        # --
        # Basic parameter
        self.task_list = task_list

        # ----
        # Changeable variable on simulation
        # --
        # Basic variables
        self.critical_path_length = (
            critical_path_length if critical_path_length != 0.0 else 0.0
        )

    def __str__(self):
        """str.

        Returns:
            str: name list of BaseTask
        Examples:
            >>> w = BaseWorkflow([BaseTask('t1')])
            >>> print([t.name for t in w.task_list])
            ['t1']
        """
        return "{}".format(list(map(lambda task: str(task), self.task_list)))

    def export_dict_json_data(self):
        """
        Export the information of this workflow to JSON data.

        Returns:
            dict: JSON format data.
        """
        dict_json_data = {}
        dict_json_data.update(
            type="BaseWorkflow",
            task_list=[t.export_dict_json_data() for t in self.task_list],
            critical_path_length=self.critical_path_length,
        )
        return dict_json_data

    def read_json_data(self, json_data):
        """
        Read the JSON data for creating BaseOrganization instance.

        Args:
            json_data (dict): JSON data.
        """
        j_list = json_data["task_list"]
        self.task_list = [
            BaseTask(
                name=j["name"],
                ID=j["ID"],
                default_work_amount=j["default_work_amount"],
                input_task_list=j["input_task_list"],
                output_task_list=j["output_task_list"],
                allocated_team_list=j["allocated_team_list"],
                allocated_workplace_list=j["allocated_workplace_list"],
                need_facility=j["need_facility"],
                target_component=j["target_component"],
                default_progress=j["default_progress"],
                due_time=j["due_time"],
                auto_task=j["auto_task"],
                fixing_allocating_worker_id_list=j["fixing_allocating_worker_id_list"],
                fixing_allocating_facility_id_list=j[
                    "fixing_allocating_facility_id_list"
                ],
                # Basic variables
                est=j["est"],
                eft=j["eft"],
                lst=j["lst"],
                lft=j["lft"],
                remaining_work_amount=j["remaining_work_amount"],
                state=BaseTaskState(j["state"]),
                state_record_list=[
                    BaseTaskState(num) for num in j["state_record_list"]
                ],
                allocated_worker_list=j["allocated_worker_list"],
                allocated_worker_id_record=j["allocated_worker_id_record"],
                allocated_facility_list=j["allocated_facility_list"],
                allocated_facility_id_record=j["allocated_facility_id_record"],
            )
            for j in j_list
        ]

        self.critical_path_length = json_data["critical_path_length"]

    def extract_none_task_list(self, target_time_list):
        """
        Extract NONE task list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract none task from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
        Returns:
            List[BaseTask]: List of BaseTask
        """
        return self.__extract_state_task_list(target_time_list, BaseTaskState.NONE)

    def extract_ready_task_list(self, target_time_list):
        """
        Extract READY task list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract ready task from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
        Returns:
            List[BaseTask]: List of BaseTask
        """
        return self.__extract_state_task_list(target_time_list, BaseTaskState.READY)

    def extract_working_task_list(self, target_time_list):
        """
        Extract WORKING task list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract working task from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
        Returns:
            List[BaseTask]: List of BaseTask
        """
        return self.__extract_state_task_list(target_time_list, BaseTaskState.WORKING)

    def extract_finished_task_list(self, target_time_list):
        """
        Extract FINISHED task list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract finished task from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
        Returns:
            List[BaseTask]: List of BaseTask
        """
        return self.__extract_state_task_list(target_time_list, BaseTaskState.FINISHED)

    def __extract_state_task_list(self, target_time_list, target_state):
        """
        Extract state task list from simulation result.

        Args:
            target_time_list (List[int]):
                Target time list.
                If you want to extract target_state task from time 2 to time 4,
                you must set [2, 3, 4] to this argument.
            target_state (BaseTaskState):
                Target state.
        Returns:
            List[BaseTask]: List of BaseTask
        """
        task_list = []
        for task in self.task_list:
            extract_flag = True
            for time in target_time_list:
                if len(task.state_record_list) <= time:
                    extract_flag = False
                    break
                if task.state_record_list[time] != target_state:
                    extract_flag = False
                    break
            if extract_flag:
                task_list.append(task)
        return task_list

    def get_task_list(
        self,
        # Basic parameters
        name=None,
        ID=None,
        default_work_amount=None,
        input_task_list=None,
        output_task_list=None,
        allocated_team_list=None,
        allocated_workplace_list=None,
        need_facility=None,
        target_component=None,
        default_progress=None,
        due_time=None,
        auto_task=None,
        fixing_allocating_worker_id_list=None,
        fixing_allocating_facility_id_list=None,
        # search param
        est=None,
        eft=None,
        lst=None,
        lft=None,
        remaining_work_amount=None,
        state=None,
        allocated_worker_list=None,
        allocated_worker_id_record=None,
        allocated_facility_list=None,
        allocated_facility_id_record=None,
    ):
        """
        Get task list by using search conditions related to BaseTask parameter.

        If there is no searching condition, this function returns all `task_list`

        Args:
            name (str, optional):
                Target task name.
                Default to None.
            ID (str, optional):
                Target task ID.
                Defaults to None.
            default_work_amount (float, optional):
                Target task default_work_amount
                Defaults to None.
            input_task_list (List[BaseTask,BaseTaskDependency], optional):
                Target task input_task_list
                Defaults to None.
            output_task_list (List[BaseTask,BaseTaskDependency], optional):
                Target task output_task_list
                Defaults to None.
            allocated_team_list (List[BaseTeam], optional):
                Target task allocated_team_list
                Defaults to None.
            allocated_workplace_list (List[BaseWorkplace], optional):
                Target task allocated_workplace_list
                Defaults to None.
            need_facility (bool, optional):
                Target task need_facility
                Defaults to None.
            target_component (BaseComponent, optional):
                Target task target_component
                Defaults to None.
            default_progress (float, optional):
                Target task default_progress
                Defaults to None.
            due_time (int, optional):
                Target task due_time
                Defaults to None.
            auto_task (bool, optional):
                Target task auto_task
                Defaults to None.
            fixing_allocating_worker_id_list (List[str], optional):
                Target task fixing_allocating_worker_id_list
                Defaults to None.
            fixing_allocating_facility_id_list (List[str], optional):
                Target task fixing_allocating_facility_id_list
                Defaults to None.
            est (float, optional):
                Target task est
                Defaults to None.
            eft (float, optional):
                Target task eft
                Defaults to None.
            lst (float, optional):
                Tainput_task lst
                Defaults to None.
            lft (float, optional):
                Target task lft
                Defaults to None.
            remaining_work_amount (float, optional):
                Target task remaining_work_amount
                Defaults to None.
            state (BaseTaskState, optional):
                Target task state
                Defaults to None.
            allocated_worker_list (List[BaseWorker], optional):
                Target task allocated_worker_list
                Defaults to None.
            allocated_worker_id_record (List[List[str]], optional):
                Target task allocated_worker_id_record
                Defaults to None.
            allocated_facility_list (List[BaseFacility], optional):
                Target task allocated_facility_list
                Defaults to None.
            allocated_facility_id_record (List[List[str]], optional):
                Target task allocated_facility_id_record
                Defaults to None.
        Returns:
            List[BaseTask]: List of BaseTask
        """
        task_list = self.task_list
        if name is not None:
            task_list = list(filter(lambda task: task.name == name, task_list))
        if ID is not None:
            task_list = list(filter(lambda task: task.ID == ID, task_list))
        if default_work_amount is not None:
            task_list = list(
                filter(
                    lambda task: task.default_work_amount == default_work_amount,
                    task_list,
                )
            )
        if input_task_list is not None:
            task_list = list(
                filter(lambda task: task.input_task_list == input_task_list, task_list)
            )
        if output_task_list is not None:
            task_list = list(
                filter(
                    lambda task: task.output_task_list == output_task_list, task_list
                )
            )
        if allocated_team_list is not None:
            task_list = list(
                filter(
                    lambda task: task.allocated_team_list == allocated_team_list,
                    task_list,
                )
            )
        if allocated_workplace_list is not None:
            task_list = list(
                filter(
                    lambda task: task.allocated_workplace_list
                    == allocated_workplace_list,
                    task_list,
                )
            )
        if need_facility is not None:
            task_list = list(
                filter(lambda task: task.need_facility == need_facility, task_list)
            )
        if target_component is not None:
            task_list = list(
                filter(
                    lambda task: task.target_component == target_component, task_list
                )
            )
        if default_progress is not None:
            task_list = list(
                filter(
                    lambda task: task.default_progress == default_progress, task_list
                )
            )
        if due_time is not None:
            task_list = list(filter(lambda task: task.due_time == due_time, task_list))
        if auto_task is not None:
            task_list = list(
                filter(lambda task: task.auto_task == auto_task, task_list)
            )
        if fixing_allocating_worker_id_list is not None:
            task_list = list(
                filter(
                    lambda task: task.fixing_allocating_worker_id_list
                    == fixing_allocating_worker_id_list,
                    task_list,
                )
            )
        if fixing_allocating_facility_id_list is not None:
            task_list = list(
                filter(
                    lambda task: task.fixing_allocating_facility_id_list
                    == fixing_allocating_facility_id_list,
                    task_list,
                )
            )
        if est is not None:
            task_list = list(filter(lambda task: task.est == est, task_list))
        if eft is not None:
            task_list = list(filter(lambda task: task.eft == eft, task_list))
        if lst is not None:
            task_list = list(filter(lambda task: task.lst == lst, task_list))
        if lft is not None:
            task_list = list(filter(lambda task: task.lft == lft, task_list))
        if remaining_work_amount is not None:
            task_list = list(
                filter(
                    lambda task: task.remaining_work_amount == remaining_work_amount,
                    task_list,
                )
            )
        if state is not None:
            task_list = list(filter(lambda task: task.state == state, task_list))
        if allocated_worker_list is not None:
            task_list = list(
                filter(
                    lambda task: task.allocated_worker_list == allocated_worker_list,
                    task_list,
                )
            )
        if allocated_worker_id_record is not None:
            task_list = list(
                filter(
                    lambda task: task.allocated_worker_id_record
                    == allocated_worker_id_record,
                    task_list,
                )
            )
        if allocated_facility_list is not None:
            task_list = list(
                filter(
                    lambda task: task.allocated_facility_list
                    == allocated_facility_list,
                    task_list,
                )
            )
        if allocated_facility_id_record is not None:
            task_list = list(
                filter(
                    lambda task: task.allocated_facility_id_record
                    == allocated_facility_id_record,
                    task_list,
                )
            )
        return task_list

    def initialize(self, state_info=True, log_info=True):
        """
        Initialize the changeable variables of BaseWorkflow including PERT calculation.

        If `state_info` is True, the following attributes are initialized.

          - `critical_path_length`
          - PERT data
          - The state of each task after all tasks are initialized.

        BaseTask in `task_list` are also initialized by this function.

        Args:
            state_info (bool):
                State information are initialized or not.
                Defaluts to True.
            log_info (bool):
                Log information are initialized or not.
                Defaults to True.
        """
        for task in self.task_list:
            task.initialize(state_info=state_info, log_info=log_info)
            if task.parent_workflow is None:
                task.parent_workflow = self
        if state_info:
            self.critical_path_length = 0.0
            self.update_PERT_data(0)
            self.check_state(-1, BaseTaskState.READY)

    def reverse_log_information(self, delete_head=False):
        """Reverse log information of all."""
        for t in self.task_list:
            t.reverse_log_information(delete_head)

    def record(self):
        """Record the state of all tasks in `task_list`."""
        for task in self.task_list:
            task.record_allocated_workers_facilities_id()
            task.record_state()

    def update_PERT_data(self, time: int):
        """
        Update PERT data (est,eft,lst,lft) of each BaseTask in task_list.

        Args:
            time (int):
                Simulation time.
        """
        self.__set_est_eft_data(time)
        self.__set_lst_lft_criticalpath_data(time)

    def check_state(self, time: int, state: BaseTaskState):
        """
        Check state of all BaseTasks in task_list.

        Args:
            time (int):
                Simulation time
            state (BaseTaskState):
                Check target state.
                Search and update all tasks which can change only target state.
        """
        if state == BaseTaskState.READY:
            self.__check_ready(time)
        elif state == BaseTaskState.WORKING:
            self.__check_working(time)
        elif state == BaseTaskState.FINISHED:
            self.__check_finished(time)

    def __check_ready(self, time: int):
        none_task_list = list(
            filter(lambda task: task.state == BaseTaskState.NONE, self.task_list)
        )
        for none_task in none_task_list:
            input_task_list = none_task.input_task_list

            # check READY condition by each dependency
            # FS: if input task is finished
            # SS: if input task is started
            # ...or this is head task
            ready = True
            for input_task, dependency in input_task_list:
                if dependency == BaseTaskDependency.FS:
                    if input_task.state == BaseTaskState.FINISHED:
                        ready = True
                    else:
                        ready = False
                        break
                elif dependency == BaseTaskDependency.SS:
                    if input_task.state == BaseTaskState.WORKING:
                        ready = True
                    else:
                        ready = False
                        break
                elif dependency == BaseTaskDependency.SF:
                    pass
                elif dependency == BaseTaskDependency.FF:
                    pass
            if ready:
                none_task.state = BaseTaskState.READY

    def __check_working(self, time: int):
        ready_and_assigned_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.READY
                and len(task.allocated_worker_list) > 0,
                self.task_list,
            )
        )

        ready_auto_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.READY and task.auto_task,
                self.task_list,
            )
        )

        working_and_assigned_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.WORKING
                and len(task.allocated_worker_list) > 0,
                self.task_list,
            )
        )

        target_task_list = []
        target_task_list.extend(ready_and_assigned_task_list)
        target_task_list.extend(ready_auto_task_list)
        target_task_list.extend(working_and_assigned_task_list)

        for task in target_task_list:
            if task.state == BaseTaskState.READY:
                task.state = BaseTaskState.WORKING
                for worker in task.allocated_worker_list:
                    worker.state = BaseWorkerState.WORKING
                    # worker.assigned_task_list.append(task)
                if task.need_facility:
                    for facility in task.allocated_facility_list:
                        facility.state = BaseFacilityState.WORKING
                        # facility.assigned_task_list.append(task)

            elif task.state == BaseTaskState.WORKING:
                for worker in task.allocated_worker_list:
                    if worker.state == BaseWorkerState.FREE:
                        worker.state = BaseWorkerState.WORKING
                        # worker.assigned_task_list.append(task)
                    if task.need_facility:
                        for facility in task.allocated_facility_list:
                            if facility.state == BaseFacilityState.FREE:
                                facility.state = BaseFacilityState.WORKING
                                # facility.assigned_task_list.append(task)

    def __check_finished(self, time: int, error_tol=1e-10):
        working_and_zero_task_list = list(
            filter(
                lambda task: task.state == BaseTaskState.WORKING
                and task.remaining_work_amount < 0.0 + error_tol,
                self.task_list,
            )
        )
        for task in working_and_zero_task_list:
            # check FINISH condition by each dependency
            # SF: if input task is working
            # FF: if input task is finished
            finished = True
            for input_task, dependency in task.input_task_list:
                if dependency == BaseTaskDependency.FS:
                    pass
                elif dependency == BaseTaskDependency.SS:
                    pass
                elif dependency == BaseTaskDependency.SF:
                    if input_task.state == BaseTaskState.WORKING:
                        finished = True
                    else:
                        finished = False
                        break
                elif dependency == BaseTaskDependency.FF:
                    if input_task.state == BaseTaskState.FINISHED:
                        finished = True
                    else:
                        finished = False
                        break
            if finished:
                task.state = BaseTaskState.FINISHED
                task.remaining_work_amount = 0.0

                for worker in task.allocated_worker_list:
                    if all(
                        list(
                            map(
                                lambda task: task.state == BaseTaskState.FINISHED,
                                worker.assigned_task_list,
                            )
                        )
                    ):
                        worker.state = BaseWorkerState.FREE
                        worker.assigned_task_list.remove(task)
                task.allocated_worker_list = []

                if task.need_facility:
                    for facility in task.allocated_facility_list:
                        if all(
                            list(
                                map(
                                    lambda task: task.state == BaseTaskState.FINISHED,
                                    facility.assigned_task_list,
                                )
                            )
                        ):
                            facility.state = BaseFacilityState.FREE
                            facility.assigned_task_list.remove(task)

                    task.allocated_facility_list = []

    def __set_est_eft_data(self, time: int):

        input_task_list = []

        # 1. Set the earliest finish time of head tasks.
        for task in self.task_list:
            task.est = time
            if len(task.input_task_list) == 0:
                task.eft = time + task.remaining_work_amount
                input_task_list.append(task)

        # 2. Calculate PERT information of all tasks
        while len(input_task_list) > 0:
            next_task_list = []
            for input_task in input_task_list:
                for next_task, dependency in input_task.output_task_list:
                    pre_est = next_task.est
                    est = 0
                    eft = 0
                    if dependency == BaseTaskDependency.FS:
                        est = input_task.est + input_task.remaining_work_amount
                        eft = est + next_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.SS:
                        est = input_task.est + 0
                        eft = est + next_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.FF:
                        est = input_task.est + 0
                        eft = est + next_task.remaining_work_amount
                        if input_task.eft > eft:
                            eft = input_task.eft
                    elif dependency == BaseTaskDependency.SF:
                        est = input_task.est + 0
                        eft = est + next_task.remaining_work_amount
                        if input_task.est > eft:
                            eft = input_task.est
                    else:
                        est = input_task.est + input_task.remaining_work_amount
                        eft = est + next_task.remaining_work_amount
                    if est >= pre_est:
                        next_task.est = est
                        next_task.eft = eft
                    next_task_list.append(next_task)

            input_task_list = next_task_list

    def __set_lst_lft_criticalpath_data(self, time: int):

        # 1. Extract the list of tail tasks.
        output_task_list = list(
            filter(lambda task: len(task.output_task_list) == 0, self.task_list)
        )

        # 2. Update the information of critical path of this workflow.
        self.critical_path_length = max(output_task_list, key=lambda task: task.eft).eft
        for task in output_task_list:
            task.lft = self.critical_path_length
            task.lst = task.lft - task.remaining_work_amount

        # 3. Calculate PERT information of all tasks
        while len(output_task_list) > 0:

            prev_task_list = []
            for output_task in output_task_list:
                for prev_task, dependency in output_task.input_task_list:
                    pre_lft = prev_task.lft
                    lst = 0
                    lft = 0
                    if dependency == BaseTaskDependency.FS:
                        lft = output_task.lst
                        lst = lft - prev_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.SS:
                        lst = output_task.lst
                        lft = lst + prev_task.remaining_work_amount
                    elif dependency == BaseTaskDependency.FF:
                        lst = output_task.lst
                        lft = lst + prev_task.remaining_work_amount
                        if output_task.lft < lft:
                            lft = output_task.lft
                    elif dependency == BaseTaskDependency.SF:
                        lst = output_task.lst
                        lft = lst + prev_task.remaining_work_amount
                        if output_task.lft < lst:
                            lst = output_task.lft
                    else:
                        lft = output_task.lst
                        lst = lft - prev_task.remaining_work_amount
                    if pre_lft < 0 or pre_lft >= lft:
                        prev_task.lst = lst
                        prev_task.lft = lft
                    prev_task_list.append(prev_task)

            output_task_list = prev_task_list

    def reverse_dependencies(self):
        """
        Reverse all task dependencies in task_list.

        Note:
            This method is developed only for backward simulation.
        """
        # 1.
        # Register the input_task_list to dummy_output_task_list
        # Register the output_task_list to dummy_input_task_list
        for task in self.task_list:
            task.dummy_output_task_list = task.input_task_list
            task.dummy_input_task_list = task.output_task_list

        # 2.
        # Register the dummy_output_task_list to output_task_list
        # Register the dummy_input_task_list to input_task_list
        # Delete the dummy_output_task_list, dummy_input_task_list
        for task in self.task_list:
            task.output_task_list = task.dummy_output_task_list
            task.input_task_list = task.dummy_input_task_list
            del task.dummy_output_task_list, task.dummy_input_task_list

    def perform(self, time: int, seed=None):
        """
        Perform BaseTask in task_list in simulation.

        Args:
            time (int):
                Simulation time.
            seed (int, optional):
                Random seed for describing deviation of progress.
                If workamount
                Defaults to None.
        """
        for task in self.task_list:
            task.perform(time, seed=seed)

    def plot_simple_gantt(
        self,
        finish_margin=1.0,
        view_auto_task=False,
        view_ready=False,
        task_color="#00EE00",
        auto_task_color="#005500",
        ready_color="#C0C0C0",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
    ):
        """
        Plot Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_auto_task (bool, optional):
                View auto_task or not.
                Defaults to False.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
            task_color (str, optional):
                Task color setting information.
                Defaults to "#00EE00".
            auto_task_color (str, optional):
                Auto Task color setting information.
                Defaults to "#005500".
            ready_color (str, optional):
                Ready Task color setting information.
                Defaults to "#C0C0C0".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
        """
        fig, gnt = self.create_simple_gantt(
            finish_margin=finish_margin,
            view_auto_task=view_auto_task,
            view_ready=view_ready,
            task_color=task_color,
            auto_task_color=auto_task_color,
            ready_color=ready_color,
            figsize=figsize,
            dpi=dpi,
            save_fig_path=save_fig_path,
        )
        return fig

    def create_simple_gantt(
        self,
        finish_margin=1.0,
        view_auto_task=False,
        view_ready=False,
        task_color="#00EE00",
        auto_task_color="#005500",
        ready_color="#C0C0C0",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
    ):
        """
        Create Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_auto_task (bool, optional):
                View auto_task or not.
                Defaults to False.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
            task_color (str, optional):
                Task color setting information.
                Defaults to "#00EE00".
            auto_task_color (str, optional):
                Auto Task color setting information.
                Defaults to "#005500".
            ready_color (str, optional):
                Ready Task color setting information.
                Defaults to "#C0C0C0".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
            gnt: ax in plt.subplots()
        """
        fig, gnt = plt.subplots()
        fig.figsize = figsize
        fig.dpi = dpi
        gnt.set_xlabel("step")
        gnt.grid(True)

        target_task_list = list(filter(lambda task: not task.auto_task, self.task_list))
        if view_auto_task:
            target_task_list = self.task_list

        yticks = [10 * (n + 1) for n in range(len(target_task_list))]
        yticklabels = [task.name for task in target_task_list]

        gnt.set_yticks(yticks)
        gnt.set_yticklabels(yticklabels)

        for ttime in range(len(target_task_list)):
            task = target_task_list[ttime]
            (
                ready_time_list,
                working_time_list,
            ) = task.get_time_list_for_gannt_chart(finish_margin=finish_margin)

            if view_ready:
                gnt.broken_barh(
                    ready_time_list, (yticks[ttime] - 5, 9), facecolors=(ready_color)
                )
            if task.auto_task:
                gnt.broken_barh(
                    working_time_list,
                    (yticks[ttime] - 5, 9),
                    facecolors=(auto_task_color),
                )
            else:
                gnt.broken_barh(
                    working_time_list, (yticks[ttime] - 5, 9), facecolors=(task_color)
                )

        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig, gnt

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
        view_ready=False,
    ):
        """
        Create data for gantt plotly from task_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.

        Returns:
            list[dict]: Gantt plotly information of this BaseWorkflow
        """
        df = []
        for task in self.task_list:
            df.extend(
                task.create_data_for_gantt_plotly(
                    init_datetime,
                    unit_timedelta,
                    finish_margin=finish_margin,
                    view_ready=view_ready,
                )
            )
        return df

    def create_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        finish_margin=1.0,
        view_ready=False,
        save_fig_path=None,
    ):
        """
        Create Gantt chart by plotly.

        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None ->
                dict(WORKING="rgb(146, 237, 5)", READY="rgb(107, 127, 135)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "State".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        colors = (
            colors
            if colors is not None
            else {"WORKING": "rgb(146, 237, 5)", "READY": "rgb(107, 127, 135)"}
        )
        index_col = index_col if index_col is not None else "State"
        df = self.create_data_for_gantt_plotly(
            init_datetime,
            unit_timedelta,
            finish_margin=finish_margin,
            view_ready=view_ready,
        )
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )

        return fig

    def get_networkx_graph(self):
        """
        Get the information of networkx graph.

        Returns:
            G: networkx.Digraph()
        """
        G = nx.DiGraph()

        # 1. add all nodes
        for task in self.task_list:
            G.add_node(task)

        # 2. add all edges
        for task in self.task_list:
            for input_task, dependency in task.input_task_list:
                G.add_edge(input_task, task)

        return G

    def draw_networkx(
        self,
        G=None,
        pos=None,
        arrows=True,
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
        **kwds,
    ):
        """
        Draw networkx.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            arrows (bool, optional):
                Digraph or Graph(no arrows).
                Defaults to True.
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.
            **kwds:
                another networkx settings.
        Returns:
            figure: Figure for a network
        """
        fig = plt.figure(figsize=figsize, dpi=dpi)
        G = G if G is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(G)

        # normal task
        normal_task_list = [task for task in self.task_list if not task.auto_task]
        nx.draw_networkx_nodes(
            G,
            pos,
            nodelist=normal_task_list,
            node_color=task_node_color,
            # **kwds,
        )

        # auto task
        auto_task_list = [task for task in self.task_list if task.auto_task]
        nx.draw_networkx_nodes(
            G,
            pos,
            nodelist=auto_task_list,
            node_color=auto_task_node_color,
            # **kwds,
        )

        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(G, pos)
        plt.axis("off")
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig

    def get_node_and_edge_trace_for_plotly_network(
        self,
        G=None,
        pos=None,
        node_size=20,
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
    ):
        """
        Get nodes and edges information of plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".

        Returns:
            task_node_trace: Normal Task Node information of plotly network.
            auto_task_node_trace: Auto Task Node information of plotly network.
            edge_trace: Edge information of plotly network.
        """
        G = G if G is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(G)

        task_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker={
                "color": task_node_color,
                "size": node_size,
            },
        )

        auto_task_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker={
                "color": auto_task_node_color,
                "size": node_size,
            },
        )

        for node in G.nodes:
            x, y = pos[node]
            if not node.auto_task:
                task_node_trace["x"] = task_node_trace["x"] + (x,)
                task_node_trace["y"] = task_node_trace["y"] + (y,)
                task_node_trace["text"] = task_node_trace["text"] + (node,)
            elif node.auto_task:
                auto_task_node_trace["x"] = auto_task_node_trace["x"] + (x,)
                auto_task_node_trace["y"] = auto_task_node_trace["y"] + (y,)
                auto_task_node_trace["text"] = auto_task_node_trace["text"] + (node,)

        edge_trace = go.Scatter(
            x=[],
            y=[],
            line={"width": 1, "color": "#888"},
            hoverinfo="none",
            mode="lines",
        )

        for edge in G.edges:
            x = edge[0]
            y = edge[1]
            xposx, xposy = pos[x]
            yposx, yposy = pos[y]
            edge_trace["x"] += (xposx, yposx)
            edge_trace["y"] += (xposy, yposy)

        return task_node_trace, auto_task_node_trace, edge_trace

    def draw_plotly_network(
        self,
        G=None,
        pos=None,
        title="Workflow",
        node_size=20,
        task_node_color="#00EE00",
        auto_task_node_color="#005500",
        save_fig_path=None,
    ):
        """
        Draw plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            title (str, optional):
                Figure title of this network.
                Defaults to "Workflow".
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            task_node_color (str, optional):
                Node color setting information.
                Defaults to "#00EE00".
            auto_task_node_color (str, optional):
                Node color setting information.
                Defaults to "#005500".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a network

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        G = G if G is not None else self.get_networkx_graph()
        pos = pos if pos is not None else nx.spring_layout(G)
        (
            task_node_trace,
            auto_task_node_trace,
            edge_trace,
        ) = self.get_node_and_edge_trace_for_plotly_network(
            G,
            pos,
            node_size=node_size,
            task_node_color=task_node_color,
            auto_task_node_color=auto_task_node_color,
        )
        fig = go.Figure(
            data=[edge_trace, task_node_trace, auto_task_node_trace],
            layout=go.Layout(
                title=title,
                showlegend=False,
                #         hovermode='closest',
                #         margin=dict(b=20,l=5,r=5,t=40),
                annotations=[
                    {
                        "ax": edge_trace["x"][index * 2],
                        "ay": edge_trace["y"][index * 2],
                        "axref": "x",
                        "ayref": "y",
                        "x": edge_trace["x"][index * 2 + 1],
                        "y": edge_trace["y"][index * 2 + 1],
                        "xref": "x",
                        "yref": "y",
                        "showarrow": True,
                        "arrowhead": 5,
                    }
                    for index in range(0, int(len(edge_trace["x"]) / 2))
                ],
                xaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
                yaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
            ),
        )
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )
        return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""base_organization."""

import abc
import datetime
import warnings
from typing import List

import matplotlib.pyplot as plt

import networkx as nx

import plotly.figure_factory as ff
import plotly.graph_objects as go

from .base_facility import BaseFacility, BaseFacilityState
from .base_team import BaseTeam
from .base_worker import BaseWorker, BaseWorkerState
from .base_workplace import BaseWorkplace


class BaseOrganization(object, metaclass=abc.ABCMeta):
    """BaseOrganization.

    BaseOrganization class for expressing organization in target project.
    BaseOrganization is consist of multiple BaseTeam and BaseWorkplace.
    This class will be used as template.

    Args:
        team_list (List[BaseTeam]):
            Basic parameter.
            List of BaseTeam in this organization.
        workplace_list (List[BaseWorkplace], optional):
            Basic parameter.
            List of BaseWorkplace in this organization.
            Default to None -> []
        cost_list (List[float], optional):
            Basic variable.
            History or record of this organization's cost in simulation.
            Defaults to None -> [].
    """

    def __init__(
        self,
        # Basic parameters
        team_list: List[BaseTeam],
        workplace_list=None,
        # Basic variables
        cost_list=None,
    ):
        """init."""
        # ----
        # Constraint parameters on simulation
        # --
        # Basic parameter
        self.team_list = team_list
        self.workplace_list = workplace_list if workplace_list is not None else []
        # --
        # Advanced parameter for customized simulation

        # ----
        # Changeable variables on simulation
        # --
        # Basic variables
        self.cost_list = []
        # --
        # Advanced variables for customized simulation

    def __str__(self):
        """str.

        Returns:
            str: name list of BaseTeam
        Examples:
            >>> o = BaseOrganization([BaseTeam('t1')])
            >>> print([t.name for t in o.team_list])
            ['t1']
        """
        return "{}".format(list(map(lambda team: str(team), self.team_list)))

    def export_dict_json_data(self):
        """
        Export the information of this organization to JSON data.

        Returns:
            dict: JSON format data.
        """
        dict_json_data = {}
        dict_json_data.update(
            type="BaseOrganization",
            team_list=[t.export_dict_json_data() for t in self.team_list],
            workplace_list=[t.export_dict_json_data() for t in self.workplace_list],
            cost_list=self.cost_list,
        )
        return dict_json_data

    def read_json_data(self, json_data):
        """
        Read the JSON data for creating BaseOrganization instance.

        Args:
            json_data (dict): JSON data.
        """
        self.team_list = []
        j_list = json_data["team_list"]
        for j in j_list:
            worker_list = []
            for w in j["worker_list"]:
                worker = BaseWorker(
                    name=w["name"],
                    ID=w["ID"],
                    team_id=w["team_id"],
                    cost_per_time=w["cost_per_time"],
                    solo_working=w["solo_working"],
                    workamount_skill_mean_map=w["workamount_skill_mean_map"],
                    workamount_skill_sd_map=w["workamount_skill_sd_map"],
                    facility_skill_map=w["facility_skill_map"],
                    state=BaseWorkerState(w["state"]),
                    state_record_list=[
                        BaseWorkerState(state_num)
                        for state_num in w["state_record_list"]
                    ],
                    cost_list=w["cost_list"],
                    assigned_task_list=w["assigned_task_list"],
                    assigned_task_id_record=w["assigned_task_id_record"],
                )
                worker_list.append(worker)
            self.team_list.append(
                BaseTeam(
                    name=j["name"],
                    ID=j["ID"],
                    worker_list=worker_list,
                    targeted_task_list=j["targeted_task_list"],
                    parent_team=j["parent_team"],
                    cost_list=j["cost_list"],
                )
            )

        self.workplace_list = []
        j_list = json_data["workplace_list"]
        for j in j_list:
            facility_list = []
            for w in j["facility_list"]:
                facility = BaseFacility(
                    name=w["name"],
                    ID=w["ID"],
                    workplace_id=w["workplace_id"],
                    cost_per_time=w["cost_per_time"],
                    solo_working=w["solo_working"],
                    workamount_skill_mean_map=w["workamount_skill_mean_map"],
                    workamount_skill_sd_map=w["workamount_skill_sd_map"],
                    state=BaseFacilityState(w["state"]),
                    state_record_list=[
                        BaseFacilityState(state_num)
                        for state_num in w["state_record_list"]
                    ],
                    cost_list=w["cost_list"],
                    assigned_task_list=w["assigned_task_list"],
                    assigned_task_id_record=w["assigned_task_id_record"],
                )
                facility_list.append(facility)
            self.workplace_list.append(
                BaseWorkplace(
                    name=j["name"],
                    ID=j["ID"],
                    facility_list=facility_list,
                    targeted_task_list=j["targeted_task_list"],
                    parent_workplace=j["parent_workplace"],
                    max_space_size=j["max_space_size"],
                    cost_list=j["cost_list"],
                    placed_component_list=j["placed_component_list"],
                    placed_component_id_record=j["placed_component_id_record"],
                )
            )

        self.cost_list = json_data["cost_list"]

    def get_team_list(
        self,
        name=None,
        ID=None,
        worker_list=None,
        targeted_task_list=None,
        parent_team=None,
        cost_list=None,
    ):
        """
        Get team list by using search conditions related to BaseTeam parameter.

        If there is no searching condition, this function returns all self.team_list

        Args:
            name (str, optional):
                Target team name.
                Defaults to None.
            ID (str, optional):
                Target team ID.
                Defaults to None.
            worker_list (List[BaseWorker], optional):
                Target team worker_list.
                Defaults to None.
            targeted_task_list (List[BaseTask], optional):
                Target team targeted_task_list.
                Defaults to None.
            parent_team (BaseTeam, optional):
                Target team parent_team.
                Defaults to None.
            cost_list (List[float], optional):
                Target team cost_list.
                Defaults to None.

        Returns:
            List[BaseTeam]: List of BaseTeam.
        """
        team_list = self.team_list
        if name is not None:
            team_list = list(filter(lambda x: x.name == name, team_list))
        if ID is not None:
            team_list = list(filter(lambda x: x.ID == ID, team_list))
        if worker_list is not None:
            team_list = list(filter(lambda x: x.worker_list == worker_list, team_list))
        if targeted_task_list is not None:
            team_list = list(
                filter(lambda x: x.targeted_task_list == targeted_task_list, team_list)
            )
        if parent_team is not None:
            team_list = list(filter(lambda x: x.parent_team == parent_team, team_list))
        if cost_list is not None:
            team_list = list(filter(lambda x: x.cost_list == cost_list, team_list))
        return team_list

    def get_workplace_list(
        self,
        name=None,
        ID=None,
        facility_list=None,
        targeted_task_list=None,
        parent_workplace=None,
        max_space_size=None,
        cost_list=None,
        placed_component_list=None,
        placed_component_id_record=None,
    ):
        """
        Get workplace list by using search conditions related to BaseTeam parameter.

        If there is no searching condition, this function returns all self.workplace_list

        Args:
            name (str, optional):
                Target workplace name.
                Defaults to None.
            ID (str, optional):
                Target workplace ID.
                Defaults to None.
            facility_list (List[BaseFacility], optional):
                Target workplace facility_list.
                Defaults to None.
            targeted_task_list (List[BaseTask], optional):
                Target workplace targeted_task_list.
                Defaults to None.
            parent_workplace (BaseWorkplace, optional):
                Target workplace parent_workplace.
                Defaults to None.
            max_space_size (float, optional):
                Target workplace max_space_size.
                Defaults to None.
            placed_component_list (List[BaseComponent], optional):
                Target workplace placed_component_list.
                Defaults to None.
            placed_component_id_record(List[List[str]], optional):
                Target workplace placed_component_id_record.
                Defaults to None.
            cost_list (List[float], optional):
                Target workplace cost_list.
                Defaults to None.

        Returns:
            List[BaseWorkplace]: List of BaseWorkplace.
        """
        workplace_list = self.workplace_list
        if name is not None:
            workplace_list = list(filter(lambda x: x.name == name, workplace_list))
        if ID is not None:
            workplace_list = list(filter(lambda x: x.ID == ID, workplace_list))
        if facility_list is not None:
            workplace_list = list(
                filter(lambda x: x.facility_list == facility_list, workplace_list)
            )
        if targeted_task_list is not None:
            workplace_list = list(
                filter(
                    lambda x: x.targeted_task_list == targeted_task_list, workplace_list
                )
            )
        if parent_workplace is not None:
            workplace_list = list(
                filter(lambda x: x.parent_workplace == parent_workplace, workplace_list)
            )
        if max_space_size is not None:
            workplace_list = list(
                filter(lambda x: x.max_space_size == max_space_size, workplace_list)
            )
        if placed_component_list is not None:
            workplace_list = list(
                filter(
                    lambda x: x.placed_component_list == placed_component_list,
                    workplace_list,
                )
            )
        if placed_component_id_record is not None:
            workplace_list = list(
                filter(
                    lambda x: x.placed_component_id_record
                    == placed_component_id_record,
                    workplace_list,
                )
            )
        if cost_list is not None:
            workplace_list = list(
                filter(lambda x: x.cost_list == cost_list, workplace_list)
            )
        return workplace_list

    def get_worker_list(
        self,
        name=None,
        ID=None,
        team_id=None,
        cost_per_time=None,
        solo_working=None,
        workamount_skill_mean_map=None,
        workamount_skill_sd_map=None,
        facility_skill_map=None,
        state=None,
        cost_list=None,
        assigned_task_list=None,
        assigned_task_id_record=None,
    ):
        """
        Get worker list by using search conditions related to BaseWorker parameter.

        This method just executes BaseTeam.get_worker_list() in self.team_list.

        Args:
            name (str, optional):
                Target worker name.
                Defaults to None.
            ID (str, optional):
                Target worker ID.
                Defaults to None.
            workplace_id (str, optional):
                Target worker workplace_id.
                Defaults to None.
            cost_per_time (float, optional):
                Target worker cost_per_time.
                Defaults to None.
            solo_working (bool, optional):
                Target worker solo_working.
                Defaults to None.
            workamount_skill_mean_map (Dict[str, float], optional):
                Target worker workamount_skill_mean_map.
                Defaults to None.
            workamount_skill_sd_map (Dict[str, float], optional):
                Target worker workamount_skill_sd_map.
                Defaults to None.
            facility_skill_map (Dict[str, float], optional):
                Target worker facility_skill_map.
                Defaults to None.
            state (BaseWorkerState, optional):
                Target worker state.
                Defaults to None.
            cost_list (List[float], optional):
                Target worker cost_list.
                Defaults to None.
            assigned_task_list (List[BaseTask], optional):
                Target worker assigned_task_list.
                Defaults to None.
            assigned_task_id_record (List[List[str]], optional):
                Target worker assigned_task_id_record.
                Defaults to None.

        Returns:
            List[BaseWorker]: List of BaseWorker.
        """
        worker_list = []
        for team in self.team_list:
            worker_list = worker_list + team.get_worker_list(
                name=name,
                ID=ID,
                team_id=team_id,
                cost_per_time=cost_per_time,
                solo_working=solo_working,
                workamount_skill_mean_map=workamount_skill_mean_map,
                workamount_skill_sd_map=workamount_skill_sd_map,
                facility_skill_map=facility_skill_map,
                state=state,
                cost_list=cost_list,
                assigned_task_list=assigned_task_list,
                assigned_task_id_record=assigned_task_id_record,
            )
        return worker_list

    def get_facility_list(
        self,
        name=None,
        ID=None,
        workplace_id=None,
        cost_per_time=None,
        solo_working=None,
        workamount_skill_mean_map=None,
        workamount_skill_sd_map=None,
        state=None,
        cost_list=None,
        assigned_task_list=None,
        assigned_task_id_record=None,
    ):
        """
        Get facility list by using search conditions related to BaseFacility parameter.

        This method just executes BaseTeam.get_facility_list() in self.workplace_list.

        Args:
            name (str, optional):
                Target facility name.
                Defaults to None.
            ID (str, optional):
                Target facility ID.
                Defaults to None.
            workplace_id (str, optional):
                Target facility workplace_id.
                Defaults to None.
            cost_per_time (float, optional):
                Target facility cost_per_time.
                Defaults to None.
            solo_working (bool, optional):
                Target facility solo_working.
                Defaults to None.
            workamount_skill_mean_map (Dict[str, float], optional):
                Target facility workamount_skill_mean_map.
                Defaults to None.
            workamount_skill_sd_map (Dict[str, float], optional):
                Target facility workamount_skill_sd_map.
                Defaults to None.
            state (BaseFacilityState, optional):
                Target facility state.
                Defaults to None.
            cost_list (List[float], optional):
                Target facility cost_list.
                Defaults to None.
            assigned_task_list (List[BaseTask], optional):
                Target facility assigned_task_list.
                Defaults to None.
            assigned_task_id_record (List[List[str]], optional):
                Target facility assigned_task_id_record.
                Defaults to None.

        Returns:
            List[BaseFacility]: List of BaseFacility.
        """
        facility_list = []
        for workplace in self.workplace_list:
            facility_list = facility_list + workplace.get_facility_list(
                name=name,
                ID=ID,
                workplace_id=workplace_id,
                cost_per_time=cost_per_time,
                solo_working=solo_working,
                workamount_skill_mean_map=workamount_skill_mean_map,
                workamount_skill_sd_map=workamount_skill_sd_map,
                state=state,
                cost_list=cost_list,
                assigned_task_list=assigned_task_list,
                assigned_task_id_record=assigned_task_id_record,
            )
        return facility_list

    def initialize(self, state_info=True, log_info=True):
        """
        Initialize the changeable variables of BaseOrganization.

        If `log_info` is True, the following attributes are initialized.
          - cost_list

        BaseTeam in `team_list` and BaseWorkplace in `workplace_list` are also initialized by this function.

        Args:
            state_info (bool):
                State information are initialized or not.
                Defaluts to True.
            log_info (bool):
                Log information are initialized or not.
                Defaults to True.
        """
        if log_info:
            self.cost_list = []
        for team in self.team_list:
            team.initialize(state_info=state_info, log_info=log_info)
        for workplace in self.workplace_list:
            workplace.initialize(state_info=state_info, log_info=log_info)

    def reverse_log_information(self, delete_head=False):
        """Reverse log information of all."""
        self.cost_list = self.cost_list[::-1]
        for team in self.team_list:
            team.reverse_log_information(delete_head)
        for workplace in self.workplace_list:
            workplace.reverse_log_information(delete_head)
        if delete_head:
            self.cost_list.pop(0)
            # cost_head = self.cost_list.pop(0)
            # self.cost_list.append(cost_head)  # insert

    def add_labor_cost(
        self,
        only_working=True,
        add_zero_to_all_workers=False,
        add_zero_to_all_facilities=False,
    ):
        """
        Add labor cost to teams and workers in this organization.

        Args:
            only_working (bool, optional):
                If True, add labor cost to only WORKING workers in this organization.
                If False, add labor cost to all workers in this organization.
                Defaults to True.
            add_zero_to_all_workers (bool, optional):
                If True, add 0 labor cost to all workers in this team.
                If False, calculate labor cost normally.
                Defaults to False.
            add_zero_to_all_facilities (bool, optional):
                If True, add 0 labor cost to all facilities in this team.
                If False, calculate labor cost normally.
                Defaults to False.

        Returns:
            float: Total labor cost of this team in this time.
        """
        cost_this_time = 0.0
        for team in self.team_list:
            cost_this_time += team.add_labor_cost(
                only_working=only_working,
                add_zero_to_all_workers=add_zero_to_all_workers,
            )
        for workplace in self.workplace_list:
            cost_this_time += workplace.add_labor_cost(
                only_working=only_working,
                add_zero_to_all_facilities=add_zero_to_all_facilities,
            )
        self.cost_list.append(cost_this_time)
        return cost_this_time

    def record(self):
        """Record assigned task id and component."""
        for team in self.team_list:
            team.record_assigned_task_id()
            team.record_all_worker_state()
        for workplace in self.workplace_list:
            workplace.record_assigned_task_id()
            workplace.record_placed_component_id()
            workplace.record_all_facility_state()

    def plot_simple_gantt(
        self,
        finish_margin=1.0,
        view_ready=False,
        view_workers=True,
        view_facilities=True,
        team_color="#0099FF",
        worker_color="#D9E5FF",
        workplace_color="#0099FF",
        facility_color="#D9E5FF",
        ready_color="#C0C0C0",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
    ):
        """
        Plot Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to Trstate = w.state_record_list[time]ue.
            team_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            workplace_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            ready_color (str, optional):
                Ready Worker/Facility color setting information.
                Defaults to "#C0C0C0".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
        """
        fig, gnt = self.create_simple_gantt(
            finish_margin=finish_margin,
            view_ready=view_ready,
            view_workers=view_workers,
            view_facilities=view_facilities,
            team_color=team_color,
            worker_color=worker_color,
            workplace_color=workplace_color,
            facility_color=facility_color,
            ready_color=ready_color,
            figsize=figsize,
            dpi=dpi,
            save_fig_path=save_fig_path,
        )
        return fig

    def create_simple_gantt(
        self,
        finish_margin=1.0,
        view_ready=False,
        view_workers=True,
        view_facilities=True,
        team_color="#0099FF",
        worker_color="#D9E5FF",
        workplace_color="#0099FF",
        facility_color="#D9E5FF",
        ready_color="#C0C0C0",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
    ):
        """
        Create Gantt chart by matplotlib.

        In this Gantt chart, datetime information is not included.
        This method will be used after simulation.

        Args:
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to Trstate = w.state_record_list[time]ue.
            team_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            workplace_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            ready_color (str, optional):
                Ready Worker/Facility color setting information.
                Defaults to "#C0C0C0".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            fig: fig in plt.subplots()
            gnt: ax in plt.subplots()
        """
        fig, gnt = plt.subplots()
        fig.figsize = figsize
        fig.dpi = dpi
        gnt.set_xlabel("step")
        gnt.grid(True)

        target_worker_list = []
        target_facility_list = []
        yticklabels = []
        if view_workers:
            for team in self.team_list:
                for worker in team.worker_list:
                    target_worker_list.append(worker)
                    yticklabels.append(team.name + ":" + worker.name)
        if view_facilities:
            for workplace in self.workplace_list:
                for facility in workplace.facility_list:
                    target_facility_list.append(facility)
                    yticklabels.append(workplace.name + ":" + facility.name)
        yticks = [
            10 * (n + 1)
            for n in range(len(target_worker_list) + len(target_facility_list))
        ]

        gnt.set_yticks(yticks)
        gnt.set_yticklabels(yticklabels)

        for ttime in range(len(target_worker_list)):
            w = target_worker_list[ttime]
            (
                ready_time_list,
                working_time_list,
            ) = w.get_time_list_for_gannt_chart(finish_margin=finish_margin)
            if view_ready:
                gnt.broken_barh(
                    ready_time_list,
                    (yticks[ttime] - 5, 9),
                    facecolors=(ready_color),
                )
            gnt.broken_barh(
                working_time_list,
                (yticks[ttime] - 5, 9),
                facecolors=(worker_color),
            )

        for ttime in range(len(target_facility_list)):
            w = target_facility_list[ttime]
            (
                ready_time_list,
                working_time_list,
            ) = w.get_time_list_for_gannt_chart(finish_margin)
            if view_ready:
                gnt.broken_barh(
                    ready_time_list,
                    (yticks[ttime + len(target_worker_list)] - 5, 9),
                    facecolors=(ready_color),
                )
            gnt.broken_barh(
                working_time_list,
                (yticks[ttime + len(target_worker_list)] - 5, 9),
                facecolors=(facility_color),
            )

        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig, gnt

    def create_data_for_gantt_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        finish_margin=1.0,
        view_ready=False,
    ):
        """
        Create data for gantt plotly from team_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
        Returns:
            List[dict]: Gantt plotly information of this BaseOrganization.
        """
        df = []
        for team in self.team_list:
            df.extend(
                team.create_data_for_gantt_plotly(
                    init_datetime,
                    unit_timedelta,
                    finish_margin=finish_margin,
                    view_ready=view_ready,
                )
            )
        for workplace in self.workplace_list:
            df.extend(
                workplace.create_data_for_gantt_plotly(
                    init_datetime,
                    unit_timedelta,
                    finish_margin=finish_margin,
                    view_ready=view_ready,
                )
            )
        return df

    def create_gantt_plotly(
        self,
        init_datetime,
        unit_timedelta,
        title="Gantt Chart",
        colors=None,
        index_col=None,
        showgrid_x=True,
        showgrid_y=True,
        group_tasks=True,
        show_colorbar=True,
        finish_margin=1.0,
        view_ready=False,
        save_fig_path=None,
    ):
        """
        Create Gantt chart by plotly.

        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of Gantt chart.
                Defaults to "Gantt Chart".
            colors (Dict[str, str], optional):
                Color setting of plotly Gantt chart.
                Defaults to None -> dict(WORKING="rgb(46, 137, 205)", READY="rgb(107, 127, 135)").
            index_col (str, optional):
                index_col of plotly Gantt chart.
                Defaults to None -> "Type".
            showgrid_x (bool, optional):
                showgrid_x of plotly Gantt chart.
                Defaults to True.
            showgrid_y (bool, optional):
                showgrid_y of plotly Gantt chart.
                Defaults to True.
            group_tasks (bool, optional):
                group_tasks of plotly Gantt chart.
                Defaults to True.
            show_colorbar (bool, optional):
                show_colorbar of plotly Gantt chart.
                Defaults to True.
            finish_margin (float, optional):
                Margin of finish time in Gantt chart.
                Defaults to 1.0.
            view_ready (bool, optional):
                View READY time or not.
                Defaults to False.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        colors = (
            colors
            if colors is not None
            else {"WORKING": "rgb(46, 137, 205)", "READY": "rgb(107, 127, 135)"}
        )
        index_col = index_col if index_col is not None else "State"
        df = self.create_data_for_gantt_plotly(
            init_datetime,
            unit_timedelta,
            finish_margin=finish_margin,
            view_ready=view_ready,
        )
        fig = ff.create_gantt(
            df,
            title=title,
            colors=colors,
            index_col=index_col,
            showgrid_x=showgrid_x,
            showgrid_y=showgrid_y,
            show_colorbar=show_colorbar,
            group_tasks=group_tasks,
        )
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )

        return fig

    def create_data_for_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
    ):
        """
        Create data for cost history plotly from cost_list of team_list.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation

        Returns:
            data (List[go.Bar(name, x, y)]: Information of cost history chart.
        """
        data = []
        x = [
            (init_datetime + time * unit_timedelta).strftime("%Y-%m-%d %H:%M:%S")
            for time in range(len(self.cost_list))
        ]
        for team in self.team_list:
            data.append(go.Bar(name=team.name, x=x, y=team.cost_list))
        for workplace in self.workplace_list:
            data.append(go.Bar(name=workplace.name, x=x, y=workplace.cost_list))
        return data

    def create_cost_history_plotly(
        self,
        init_datetime: datetime.datetime,
        unit_timedelta: datetime.timedelta,
        title="Cost Chart",
        save_fig_path=None,
    ):
        """
        Create cost chart by plotly.

        This method will be used after simulation.

        Args:
            init_datetime (datetime.datetime):
                Start datetime of project
            unit_timedelta (datetime.timedelta):
                Unit time of simulation
            title (str, optional):
                Title of cost chart.
                Defaults to "Cost Chart".
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a gantt chart

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        data = self.create_data_for_cost_history_plotly(init_datetime, unit_timedelta)
        fig = go.Figure(data)
        fig.update_layout(barmode="stack", title=title)
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )
        return fig

    def get_networkx_graph(self, view_workers=False, view_facilities=False):
        """
        Get the information of networkx graph.

        Args:
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
        Returns:
            G: networkx.Digraph()
        """
        G = nx.DiGraph()

        # BaseTeam
        # 1. add all nodes
        for team in self.team_list:
            G.add_node(team)

        # 2. add all edges
        for team in self.team_list:
            if team.parent_team is not None:
                G.add_edge(team.parent_team, team)

        if view_workers:
            for team in self.team_list:
                for w in team.worker_list:
                    G.add_node(w)
                    G.add_edge(team, w)

        # BaseWorkplace
        # 1. add all nodes
        for workplace in self.workplace_list:
            G.add_node(workplace)

        # 2. add all edges
        for workplace in self.workplace_list:
            if workplace.parent_workplace is not None:
                G.add_edge(workplace.parent_workplace, workplace)

        if view_facilities:
            for workplace in self.workplace_list:
                for w in workplace.facility_list:
                    G.add_node(w)
                    G.add_edge(workplace, w)

        return G

    def draw_networkx(
        self,
        G=None,
        pos=None,
        arrows=True,
        view_workers=False,
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_facilities=False,
        workplace_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        figsize=[6.4, 4.8],
        dpi=100.0,
        save_fig_path=None,
        **kwds,
    ):
        """
        Draw networkx.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            arrows (bool, optional):
                Digraph or Graph(no arrows).
                Defaults to True.
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
            workplace_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            figsize ((float, float), optional):
                Width, height in inches.
                Default to [6.4, 4.8]
            dpi (float, optional):
                The resolution of the figure in dots-per-inch.
                Default to 100.0
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.
            **kwds:
                another networkx settings.
        Returns:
            figure: Figure for a network
        """
        fig = plt.figure(figsize=figsize, dpi=dpi)
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)

        # team
        nx.draw_networkx_nodes(
            G,
            pos,
            nodelist=self.team_list,
            node_color=team_node_color,
            # **kwds,
        )
        # workers
        if view_workers:

            worker_list = []
            for team in self.team_list:
                worker_list.extend(team.worker_list)

            nx.draw_networkx_nodes(
                G,
                pos,
                nodelist=worker_list,
                node_color=worker_node_color,
                # **kwds,
            )

        # workplace
        nx.draw_networkx_nodes(
            G,
            pos,
            nodelist=self.workplace_list,
            node_color=workplace_node_color,
            # **kwds,
        )
        # facility
        if view_facilities:

            facility_list = []
            for workplace in self.workplace_list:
                facility_list.extend(workplace.facility_list)

            nx.draw_networkx_nodes(
                G,
                pos,
                nodelist=facility_list,
                node_color=facility_node_color,
                # **kwds,
            )

        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(G, pos)
        plt.axis("off")
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        plt.close()
        return fig

    def get_node_and_edge_trace_for_plotly_network(
        self,
        G=None,
        pos=None,
        node_size=20,
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        workplace_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        view_facilities=False,
    ):
        """
        Get nodes and edges information of plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            workplace_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.

        Returns:
            team_node_trace: Team Node information of plotly network.
            worker_node_trace: Worker Node information of plotly network.
            workplace_node_trace: Workplace Node information of plotly network.
            facility_node_trace: Facility Node information of plotly network.
            edge_trace: Edge information of plotly network.
        """
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)

        team_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker={
                "color": team_node_color,
                "size": node_size,
            },
        )
        worker_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker={
                "color": worker_node_color,
                "size": node_size,
            },
        )
        workplace_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker={
                "color": workplace_node_color,
                "size": node_size,
            },
        )
        facility_node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode="markers",
            hoverinfo="text",
            marker={
                "color": facility_node_color,
                "size": node_size,
            },
        )

        for node in G.nodes:
            x, y = pos[node]
            if isinstance(node, BaseTeam):
                team_node_trace["x"] = team_node_trace["x"] + (x,)
                team_node_trace["y"] = team_node_trace["y"] + (y,)
                team_node_trace["text"] = team_node_trace["text"] + (node,)
            elif isinstance(node, BaseWorkplace):
                workplace_node_trace["x"] = workplace_node_trace["x"] + (x,)
                workplace_node_trace["y"] = workplace_node_trace["y"] + (y,)
                workplace_node_trace["text"] = workplace_node_trace["text"] + (node,)
            elif isinstance(node, BaseFacility):
                facility_node_trace["x"] = facility_node_trace["x"] + (x,)
                facility_node_trace["y"] = facility_node_trace["y"] + (y,)
                facility_node_trace["text"] = facility_node_trace["text"] + (node,)
            elif isinstance(node, BaseWorker):
                worker_node_trace["x"] = worker_node_trace["x"] + (x,)
                worker_node_trace["y"] = worker_node_trace["y"] + (y,)
                worker_node_trace["text"] = worker_node_trace["text"] + (node,)

        edge_trace = go.Scatter(
            x=[],
            y=[],
            line={"width": 1, "color": "#888"},
            hoverinfo="none",
            mode="lines",
        )

        for edge in G.edges:
            x = edge[0]
            y = edge[1]
            xposx, xposy = pos[x]
            yposx, yposy = pos[y]
            edge_trace["x"] += (xposx, yposx)
            edge_trace["y"] += (xposy, yposy)

        return (
            team_node_trace,
            worker_node_trace,
            workplace_node_trace,
            facility_node_trace,
            edge_trace,
        )

    def draw_plotly_network(
        self,
        G=None,
        pos=None,
        title="Organization",
        node_size=20,
        team_node_color="#0099FF",
        worker_node_color="#D9E5FF",
        view_workers=False,
        workplace_node_color="#0099FF",
        facility_node_color="#D9E5FF",
        view_facilities=False,
        save_fig_path=None,
    ):
        """
        Draw plotly network.

        Args:
            G (networkx.Digraph, optional):
                The information of networkx graph.
                Defaults to None -> self.get_networkx_graph().
            pos (networkx.layout, optional):
                Layout of networkx.
                Defaults to None -> networkx.spring_layout(G).
            title (str, optional):
                Figure title of this network.
                Defaults to "Organization".
            node_size (int, optional):
                Node size setting information.
                Defaults to 20.
            team_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            worker_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_workers (bool, optional):
                Including workers in networkx graph or not.
                Default to False.
            workplace_node_color (str, optional):
                Node color setting information.
                Defaults to "#0099FF".
            facility_node_color (str, optional):
                Node color setting information.
                Defaults to "#D9E5FF".
            view_facilities (bool, optional):
                Including facilities in networkx graph or not.
                Default to False.
            save_fig_path (str, optional):
                Path of saving figure.
                Defaults to None.

        Returns:
            figure: Figure for a network

        TODO:
            Now, save_fig_path can be utilized only json and html format.
            Saving figure png, jpg, svg file is not implemented...
        """
        G = (
            G
            if G is not None
            else self.get_networkx_graph(
                view_workers=view_workers, view_facilities=view_facilities
            )
        )
        pos = pos if pos is not None else nx.spring_layout(G)
        (
            team_node_trace,
            worker_node_trace,
            workplace_node_trace,
            facility_node_trace,
            edge_trace,
        ) = self.get_node_and_edge_trace_for_plotly_network(
            G,
            pos,
            node_size=node_size,
            team_node_color=team_node_color,
            worker_node_color=worker_node_color,
            view_workers=view_workers,
            workplace_node_color=workplace_node_color,
            facility_node_color=facility_node_color,
            view_facilities=view_facilities,
        )
        fig = go.Figure(
            data=[
                edge_trace,
                team_node_trace,
                worker_node_trace,
                workplace_node_trace,
                facility_node_trace,
            ],
            layout=go.Layout(
                title=title,
                showlegend=False,
                #         hovermode='closest',
                #         margin=dict(b=20,l=5,r=5,t=40),
                annotations=[
                    {
                        "ax": edge_trace["x"][index * 2],
                        "ay": edge_trace["y"][index * 2],
                        "axref": "x",
                        "ayref": "y",
                        "x": edge_trace["x"][index * 2 + 1],
                        "y": edge_trace["y"][index * 2 + 1],
                        "xref": "x",
                        "yref": "y",
                        "showarrow": True,
                        "arrowhead": 5,
                    }
                    for index in range(0, int(len(edge_trace["x"]) / 2))
                ],
                xaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
                yaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
            ),
        )
        if save_fig_path is not None:
            # fig.write_image(save_fig_path)
            dot_point = save_fig_path.rfind(".")

            save_mode = "error" if dot_point == -1 else save_fig_path[dot_point + 1 :]

            if save_mode == "html":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_html(save_fig_path)
            elif save_mode == "json":
                fig_go_figure = go.Figure(fig)
                fig_go_figure.write_json(save_fig_path)
            elif save_mode in ["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps"]:
                # We need to install plotly/orca
                # and set `plotly.io.orca.config.executable = '/path/to/orca'``
                # fig_go_figure = go.Figure(fig)
                # fig_go_figure.write_html(save_fig_path)
                save_mode = "error"

            if save_mode == "error":
                warnings.warn(
                    "Sorry, the function of saving this type is not implemented now. "
                    "pDESy is only support html and json in saving plotly."
                )
        return fig
